@@ -86,27 +86,27 @@ const struct pl_color_repr pl_color_repr_unknown = {0};
 
 const struct pl_color_repr pl_color_repr_rgb = {
     .sys    = PL_COLOR_SYSTEM_RGB,
-    .levels = PL_COLOR_LEVELS_PC,
+    .levels = PL_COLOR_LEVELS_FULL,
 };
 
 const struct pl_color_repr pl_color_repr_sdtv = {
     .sys    = PL_COLOR_SYSTEM_BT_601,
-    .levels = PL_COLOR_LEVELS_TV,
+    .levels = PL_COLOR_LEVELS_LIMITED,
 };
 
 const struct pl_color_repr pl_color_repr_hdtv = {
     .sys    = PL_COLOR_SYSTEM_BT_709,
-    .levels = PL_COLOR_LEVELS_TV,
+    .levels = PL_COLOR_LEVELS_LIMITED,
 };
 
 const struct pl_color_repr pl_color_repr_uhdtv = {
     .sys    = PL_COLOR_SYSTEM_BT_2020_NC,
-    .levels = PL_COLOR_LEVELS_TV,
+    .levels = PL_COLOR_LEVELS_LIMITED,
 };
 
 const struct pl_color_repr pl_color_repr_jpeg = {
     .sys    = PL_COLOR_SYSTEM_BT_601,
-    .levels = PL_COLOR_LEVELS_PC,
+    .levels = PL_COLOR_LEVELS_FULL,
 };
 
 bool pl_color_repr_equal(const struct pl_color_repr *c1,
@@ -145,8 +145,8 @@ enum pl_color_levels pl_color_levels_guess(const struct pl_color_repr *repr)
         return repr->levels;
 
     return pl_color_system_is_ycbcr_like(repr->sys)
-                ? PL_COLOR_LEVELS_TV
-                : PL_COLOR_LEVELS_PC;
+                ? PL_COLOR_LEVELS_LIMITED
+                : PL_COLOR_LEVELS_FULL;
 }
 
 float pl_color_repr_normalize(struct pl_color_repr *repr)
@@ -159,10 +159,12 @@ float pl_color_repr_normalize(struct pl_color_repr *repr)
         bits->bit_shift = 0;
     }
 
+    // If one of these is set but not the other, use the set one
     int tex_bits = PL_DEF(bits->sample_depth, 8);
-    int col_bits = PL_DEF(bits->color_depth,  8);
+    int col_bits = PL_DEF(bits->color_depth, tex_bits);
+    tex_bits = PL_DEF(tex_bits, col_bits);
 
-    if (pl_color_levels_guess(repr) == PL_COLOR_LEVELS_TV) {
+    if (pl_color_levels_guess(repr) == PL_COLOR_LEVELS_LIMITED) {
         // Limit range is always shifted directly
         scale *= (float) (1LL << tex_bits) / (1LL << col_bits);
     } else {
@@ -182,6 +184,7 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     case PL_COLOR_PRIM_BT_601_625:
     case PL_COLOR_PRIM_BT_709:
     case PL_COLOR_PRIM_BT_470M:
+    case PL_COLOR_PRIM_EBU_3213:
         return false;
     case PL_COLOR_PRIM_BT_2020:
     case PL_COLOR_PRIM_APPLE:
@@ -192,6 +195,7 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     case PL_COLOR_PRIM_DISPLAY_P3:
     case PL_COLOR_PRIM_V_GAMUT:
     case PL_COLOR_PRIM_S_GAMUT:
+    case PL_COLOR_PRIM_FILM_C:
         return true;
     case PL_COLOR_PRIM_COUNT: break;
     }
@@ -360,11 +364,12 @@ void pl_color_space_infer(struct pl_color_space *space)
 }
 
 const struct pl_color_adjustment pl_color_adjustment_neutral = {
-    .brightness = 0.0,
-    .contrast   = 1.0,
-    .saturation = 1.0,
-    .hue        = 0.0,
-    .gamma      = 1.0,
+    .brightness     = 0.0,
+    .contrast       = 1.0,
+    .saturation     = 1.0,
+    .hue            = 0.0,
+    .gamma          = 1.0,
+    .temperature    = 0.0,
 };
 
 void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y)
@@ -400,6 +405,23 @@ void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y)
     }
 }
 
+struct pl_cie_xy pl_white_from_temp(float temp)
+{
+    temp = PL_CLAMP(temp, 2500, 25000);
+
+    double ti = 1000.0 / temp, ti2 = ti * ti, ti3 = ti2 * ti, x;
+    if (temp <= 7000) {
+        x = -4.6070 * ti3 + 2.9678 * ti2 + 0.09911 * ti + 0.244063;
+    } else {
+        x = -2.0064 * ti3 + 1.9018 * ti2 + 0.24748 * ti + 0.237040;
+    }
+
+    return (struct pl_cie_xy) {
+        .x = x,
+        .y = -3 * (x*x) + 2.87 * x - 0.275,
+    };
+}
+
 const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim)
 {
     /*
@@ -503,6 +525,19 @@ const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim
             .blue  = {0.100, -0.05},
             .white = CIE_D65,
         },
+    // From FFmpeg source code
+        [PL_COLOR_PRIM_FILM_C] = {
+            .red   = {0.681, 0.319},
+            .green = {0.243, 0.692},
+            .blue  = {0.145, 0.049},
+            .white = CIE_C,
+        },
+        [PL_COLOR_PRIM_EBU_3213] = {
+            .red   = {0.630, 0.340},
+            .green = {0.295, 0.605},
+            .blue  = {0.155, 0.077},
+            .white = CIE_D65,
+        },
     };
 
     // This is the default assumption if no colorspace information could
@@ -611,6 +646,21 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     pl_matrix3x3_mul(mat, &tmp);
 }
 
+struct pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie_xy dst)
+{
+    // Use BT.709 primaries (with chosen white point) as an XYZ reference
+    struct pl_raw_primaries csp = *pl_raw_primaries_get(PL_COLOR_PRIM_BT_709);
+    csp.white = src;
+
+    struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&csp);
+    struct pl_matrix3x3 xyz2rgb = rgb2xyz;
+    pl_matrix3x3_invert(&xyz2rgb);
+
+    apply_chromatic_adaptation(src, dst, &xyz2rgb);
+    pl_matrix3x3_mul(&xyz2rgb, &rgb2xyz);
+    return xyz2rgb;
+}
+
 const struct pl_cone_params pl_vision_normal        = {PL_CONE_NONE, 1.0};
 const struct pl_cone_params pl_vision_protanomaly   = {PL_CONE_L,    0.5};
 const struct pl_cone_params pl_vision_protanopia    = {PL_CONE_L,    0.0};
@@ -782,6 +832,36 @@ struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *s
     return xyz2rgb_d;
 }
 
+// Test the sign of 'p' relative to the line 'ab' (barycentric coordinates)
+static float test_point_line(const struct pl_cie_xy p,
+                             const struct pl_cie_xy a,
+                             const struct pl_cie_xy b)
+{
+    return (p.x - b.x) * (a.y - b.y) - (a.x - b.x) * (p.y - b.y);
+}
+
+// Test if a point is entirely inside a gamut
+static float test_point_gamut(struct pl_cie_xy point,
+                              const struct pl_raw_primaries *prim)
+{
+    float d1 = test_point_line(point, prim->red, prim->green),
+          d2 = test_point_line(point, prim->green, prim->blue),
+          d3 = test_point_line(point, prim->blue, prim->red);
+
+    bool has_neg = d1 < 0 || d2 < 0 || d3 < 0,
+         has_pos = d1 > 0 || d2 > 0 || d3 > 0;
+
+    return !(has_neg && has_pos);
+}
+
+bool pl_primaries_superset(const struct pl_raw_primaries *a,
+                           const struct pl_raw_primaries *b)
+{
+    return test_point_gamut(b->red, a) &&
+           test_point_gamut(b->green, a) &&
+           test_point_gamut(b->blue, a);
+}
+
 /* Fill in the Y, U, V vectors of a yuv-to-rgb conversion matrix
  * based on the given luma weights of the R, G and B components (lr, lg, lb).
  * lr+lg+lb is assumed to equal 1.
@@ -812,6 +892,21 @@ static struct pl_matrix3x3 luma_coeffs(float lr, float lg, float lb)
     }};
 }
 
+// Applies hue and saturation controls to a YCbCr->RGB matrix
+static inline void apply_hue_sat(struct pl_matrix3x3 *m,
+                                 const struct pl_color_adjustment *params)
+{
+    // Hue is equivalent to rotating input [U, V] subvector around the origin.
+    // Saturation scales [U, V].
+    float huecos = params->saturation * cos(params->hue);
+    float huesin = params->saturation * sin(params->hue);
+    for (int i = 0; i < 3; i++) {
+        float u = m->m[i][1], v = m->m[i][2];
+        m->m[i][1] = huecos * u - huesin * v;
+        m->m[i][2] = huesin * u + huecos * v;
+    }
+}
+
 struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
                                     const struct pl_color_adjustment *params)
 {
@@ -831,8 +926,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
             {0, 1, 0}
         }};
         break;
-    case PL_COLOR_SYSTEM_BT_2100_PQ:
-    case PL_COLOR_SYSTEM_BT_2100_HLG: {
+    case PL_COLOR_SYSTEM_BT_2100_PQ: {
         // Reversed from the matrix in the spec, hard-coded for efficiency
         // and precision reasons. Exact values truncated from ITU-T H-series
         // Supplement 18.
@@ -844,6 +938,16 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         }};
         break;
     }
+    case PL_COLOR_SYSTEM_BT_2100_HLG: {
+        // Similar to BT.2100 PQ, exact values truncated from WolframAlpha
+        static const float lm_t = 0.01571858011, lm_p = 0.2095810681;
+        m = (struct pl_matrix3x3) {{
+            {1.0,  lm_t,  lm_p},
+            {1.0, -lm_t, -lm_p},
+            {1.0, 1.02127108, -0.605274491},
+        }};
+        break;
+    }
     case PL_COLOR_SYSTEM_YCGCO:
         m = (struct pl_matrix3x3) {{
             {1,  -1,  1},
@@ -864,22 +968,29 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     default: abort();
     }
 
-    struct pl_transform3x3 out = { .mat = m };
-
     // Apply hue and saturation in the correct way depending on the colorspace.
     if (pl_color_system_is_ycbcr_like(repr->sys)) {
-        // Hue is equivalent to rotating input [U, V] subvector around the origin.
-        // Saturation scales [U, V].
-        float huecos = params->saturation * cos(params->hue);
-        float huesin = params->saturation * sin(params->hue);
-        for (int i = 0; i < 3; i++) {
-            float u = out.mat.m[i][1], v = out.mat.m[i][2];
-            out.mat.m[i][1] = huecos * u - huesin * v;
-            out.mat.m[i][2] = huesin * u + huecos * v;
-        }
+        apply_hue_sat(&m, params);
+    } else if (params->saturation != 1.0 || params->hue != 0.0) {
+        // Arbitrarily simulate hue shifts using the BT.709 YCbCr model
+        struct pl_matrix3x3 yuv2rgb = luma_coeffs(0.2126, 0.7152, 0.0722);
+        struct pl_matrix3x3 rgb2yuv = yuv2rgb;
+        pl_matrix3x3_invert(&rgb2yuv);
+        apply_hue_sat(&yuv2rgb, params);
+        // M := RGB<-YUV * YUV<-RGB * M
+        pl_matrix3x3_rmul(&rgb2yuv, &m);
+        pl_matrix3x3_rmul(&yuv2rgb, &m);
     }
-    // FIXME: apply saturation for RGB
 
+    // Apply color temperature adaptation, relative to BT.709 primaries
+    if (params->temperature) {
+        struct pl_cie_xy src = pl_white_from_temp(6500);
+        struct pl_cie_xy dst = pl_white_from_temp(6500 + 3500 * params->temperature);
+        struct pl_matrix3x3 adapt = pl_get_adaptation_matrix(src, dst);
+        pl_matrix3x3_rmul(&adapt, &m);
+    }
+
+    struct pl_transform3x3 out = { .mat = m };
     int bit_depth = PL_DEF(repr->bits.sample_depth,
                     PL_DEF(repr->bits.color_depth, 8));
 
@@ -887,14 +998,14 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     double scale = (1LL << bit_depth) / ((1LL << bit_depth) - 1.0);
 
     switch (pl_color_levels_guess(repr)) {
-    case PL_COLOR_LEVELS_TV: {
+    case PL_COLOR_LEVELS_LIMITED: {
         ymax = 235 / 256. * scale;
         ymin =  16 / 256. * scale;
         cmax = 240 / 256. * scale;
         cmid = 128 / 256. * scale;
         break;
     }
-    case PL_COLOR_LEVELS_PC:
+    case PL_COLOR_LEVELS_FULL:
         // Note: For full-range YUV, there are multiple, subtly inconsistent
         // standards. So just pick the sanest implementation, which is to
         // assume MAX_INT == 1.0.
@@ -939,7 +1050,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
 
     // Update the metadata to reflect the change.
     repr->sys    = PL_COLOR_SYSTEM_RGB;
-    repr->levels = PL_COLOR_LEVELS_PC;
+    repr->levels = PL_COLOR_LEVELS_FULL;
 
     return out;
 }
@@ -947,10 +1058,20 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
 bool pl_icc_profile_equal(const struct pl_icc_profile *p1,
                           const struct pl_icc_profile *p2)
 {
-    // Test for presence of a pointer first
-    if (!!p1->data != !!p2->data)
+    if (p1->len != p2->len)
         return false;
 
-    // Otherwise, test for equality of signature+len (if a profile is present)
-    return !p1->data || (p1->signature == p2->signature && p1->len == p2->len);
+    // Ignore signatures on length-0 profiles, as a special case
+    return !p1->len || p1->signature == p2->signature;
+}
+
+void pl_icc_profile_compute_signature(struct pl_icc_profile *profile)
+{
+    // In theory, we could get this value from the profile header itself if
+    // lcms is available, but I'm not sure if it's even worth the trouble. Just
+    // hard-code this to a siphash64(), which is decently fast anyway.
+    profile->signature = pl_str_hash((pl_str) {
+        .buf = (uint8_t *) profile->data,
+        .len = profile->len
+    });
 }
