@@ -99,31 +99,40 @@ double pl_filter_sample(const struct pl_filter_config *c, double x)
     return k < 0 ? (1 - c->clamp) * k : k;
 }
 
-// Calculate a single filter row of a 1D filter, for a given phase value /
-// subpixel offset `offset`. Writes exactly f->row_size values to *out.
+// Compute a single row of weights for a given filter in one dimension, indexed
+// by the indicated subpixel offset. Writes `f->row_size` values to `out`.
 static void compute_row(struct pl_filter *f, double offset, float *out)
 {
-    pl_assert(f->row_size > 0);
-    double sum = 0;
+    double wsum = 0.0;
     for (int i = 0; i < f->row_size; i++) {
-        double x = offset - (i - f->row_size / 2 + 1);
-        // Readjust the value range to account for a stretched kernel.
+        // For the example of a filter with row size 4 and offset 0.3, we have:
+        //
+        // 0    1 *  2    3
+        //
+        // * indicates the sampled position. What we want to compute is the
+        // distance from each index to that sampled position.
+        pl_assert(f->row_size % 2 == 0);
+        const int base = f->row_size / 2 - 1; // index to the left of the center
+        const double center = base + offset; // offset of center relative to idx 0
+        double x = i - center;
+
+        // Stretch/squish the kernel by readjusting the value range
         x *= f->params.config.kernel->radius / f->radius;
-        double weight = pl_filter_sample(&f->params.config, x);
-        out[i] = weight;
-        sum += weight;
-    }
-    // Normalize to preserve energy
-    if (sum > 0.0) {
-        for (int i = 0; i < f->row_size; i++)
-            out[i] /= sum;
+        double w = pl_filter_sample(&f->params.config, x);
+        out[i] = w;
+        wsum += w;
     }
+
+    // Readjust weights to preserve energy
+    pl_assert(wsum > 0);
+    for (int i = 0; i < f->row_size; i++)
+        out[i] /= wsum;
 }
 
-static struct pl_filter_function *dupfilter(void *tactx,
+static struct pl_filter_function *dupfilter(void *alloc,
                                             const struct pl_filter_function *f)
 {
-    return f ? talloc_memdup(tactx, (void *)f, sizeof(*f)) : NULL;
+    return f ? pl_memdup(alloc, (void *)f, sizeof(*f)) : NULL;
 }
 
 const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
@@ -135,7 +144,7 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
         return NULL;
     }
 
-    struct pl_filter *f = talloc_zero(ctx, struct pl_filter);
+    struct pl_filter *f = pl_zalloc_ptr(ctx, f);
     f->params = *params;
     f->params.config.kernel = dupfilter(f, params->config.kernel);
     f->params.config.window = dupfilter(f, params->config.window);
@@ -149,7 +158,7 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
     float *weights;
     if (params->config.polar) {
         // Compute a 1D array indexed by radius
-        weights = talloc_array(f, float, params->lut_entries);
+        weights = pl_alloc(f, params->lut_entries * sizeof(float));
         f->radius_cutoff = 0.0;
         for (int i = 0; i < params->lut_entries; i++) {
             double x = radius * i / (params->lut_entries - 1);
@@ -170,7 +179,7 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
         f->row_stride = PL_ALIGN(f->row_size, params->row_stride_align);
 
         // Compute a 2D array indexed by the subpixel position
-        weights = talloc_zero_array(f, float, params->lut_entries * f->row_stride);
+        weights = pl_calloc(f, params->lut_entries * f->row_stride, sizeof(float));
         for (int i = 0; i < params->lut_entries; i++) {
             compute_row(f, i / (double)(params->lut_entries - 1),
                         weights + f->row_stride * i);
@@ -183,30 +192,30 @@ const struct pl_filter *pl_filter_generate(struct pl_context *ctx,
 
 void pl_filter_free(const struct pl_filter **filter)
 {
-    TA_FREEP((void **) filter);
+    pl_free_ptr((void **) filter);
 }
 
-const struct pl_named_filter_function *pl_find_named_filter_function(const char *name)
+const struct pl_filter_function_preset *pl_find_filter_function_preset(const char *name)
 {
     if (!name)
         return NULL;
 
-    for (int i = 0; pl_named_filter_functions[i].function; i++) {
-        if (strcmp(pl_named_filter_functions[i].name, name) == 0)
-            return &pl_named_filter_functions[i];
+    for (int i = 0; pl_filter_function_presets[i].name; i++) {
+        if (strcmp(pl_filter_function_presets[i].name, name) == 0)
+            return &pl_filter_function_presets[i];
     }
 
     return NULL;
 }
 
-const struct pl_named_filter_config *pl_find_named_filter(const char *name)
+const struct pl_filter_preset *pl_find_filter_preset(const char *name)
 {
     if (!name)
         return NULL;
 
-    for (int i = 0; pl_named_filters[i].filter; i++) {
-        if (strcmp(pl_named_filters[i].name, name) == 0)
-            return &pl_named_filters[i];
+    for (int i = 0; pl_filter_presets[i].name; i++) {
+        if (strcmp(pl_filter_presets[i].name, name) == 0)
+            return &pl_filter_presets[i];
     }
 
     return NULL;
@@ -483,7 +492,8 @@ const struct pl_filter_function pl_filter_function_spline64 = {
 };
 
 // Named filter functions
-const struct pl_named_filter_function pl_named_filter_functions[] = {
+const struct pl_filter_function_preset pl_filter_function_presets[] = {
+    {"none",            NULL},
     {"box",             &pl_filter_function_box},
     {"dirichlet",       &pl_filter_function_box}, // alias
     {"triangle",        &pl_filter_function_triangle},
@@ -510,6 +520,8 @@ const struct pl_named_filter_function pl_named_filter_functions[] = {
     {0},
 };
 
+const int pl_num_filter_function_presets = PL_ARRAY_SIZE(pl_filter_function_presets) - 1;
+
 // Built-in filter function presets
 const struct pl_filter_config pl_filter_spline16 = {
     .kernel = &pl_filter_function_spline16,
@@ -523,11 +535,11 @@ const struct pl_filter_config pl_filter_spline64 = {
     .kernel = &pl_filter_function_spline64,
 };
 
-const struct pl_filter_config pl_filter_box = {
+const struct pl_filter_config pl_filter_nearest = {
     .kernel = &pl_filter_function_box,
 };
 
-const struct pl_filter_config pl_filter_triangle = {
+const struct pl_filter_config pl_filter_bilinear = {
     .kernel = &pl_filter_function_triangle,
 };
 
@@ -586,16 +598,6 @@ const struct pl_filter_config pl_filter_ewa_hann = {
     .polar = true,
 };
 
-const struct pl_filter_config pl_filter_haasnsoft = {
-    .kernel = &jinc3,
-    .window = &pl_filter_function_hann,
-    // The blur is tuned to equal out orthogonal and diagonal contributions
-    // on a regular grid. This has the effect of almost completely killing
-    // aliasing.
-    .blur = 1.11,
-    .polar = true,
-};
-
 // Spline family
 const struct pl_filter_config pl_filter_bicubic = {
     .kernel = &pl_filter_function_bicubic,
@@ -609,6 +611,11 @@ const struct pl_filter_config pl_filter_mitchell = {
     .kernel = &pl_filter_function_mitchell,
 };
 
+const struct pl_filter_config pl_filter_mitchell_clamp = {
+    .kernel = &pl_filter_function_mitchell,
+    .clamp = 1.0,
+};
+
 const struct pl_filter_config pl_filter_robidoux = {
     .kernel = &pl_filter_function_robidoux,
 };
@@ -628,30 +635,37 @@ const struct pl_filter_config pl_filter_ewa_robidouxsharp = {
 };
 
 // Named filter configs
-const struct pl_named_filter_config pl_named_filters[] = {
-    {"spline16",            &pl_filter_spline16},
-    {"spline36",            &pl_filter_spline36},
-    {"spline64",            &pl_filter_spline64},
-    {"box",                 &pl_filter_box},
-    {"nearest",             &pl_filter_box}, // alias
-    {"triangle",            &pl_filter_triangle},
-    {"bilinear",            &pl_filter_triangle}, // alias
-    {"gaussian",            &pl_filter_gaussian},
-    {"sinc",                &pl_filter_sinc},
-    {"lanczos",             &pl_filter_lanczos},
-    {"ginseng",             &pl_filter_ginseng},
-    {"ewa_jinc",            &pl_filter_ewa_jinc},
-    {"ewa_lanczos",         &pl_filter_ewa_lanczos},
-    {"ewa_ginseng",         &pl_filter_ewa_ginseng},
-    {"ewa_hann",            &pl_filter_ewa_hann},
+const struct pl_filter_preset pl_filter_presets[] = {
+    // Highest priority / recommended filters
+    {"none",                NULL,                   "Built-in sampling"},
+    {"bilinear",            &pl_filter_bilinear,    "Bilinear"},
+    {"nearest",             &pl_filter_nearest,     "Nearest neighbour"},
+    {"bicubic",             &pl_filter_bicubic,     "Bicubic"},
+    {"lanczos",             &pl_filter_lanczos,     "Lanczos"},
+    {"ewa_lanczos",         &pl_filter_ewa_lanczos, "Jinc (EWA Lanczos)"},
+    {"gaussian",            &pl_filter_gaussian,    "Gaussian"},
+    {"spline16",            &pl_filter_spline16,    "Spline (2 taps)"},
+    {"spline36",            &pl_filter_spline36,    "Spline (3 taps)"},
+    {"spline64",            &pl_filter_spline64,    "Spline (4 taps)"},
+    {"mitchell",            &pl_filter_mitchell,    "Mitchell-Netravali"},
+
+    // Remaining filters
+    {"sinc",                &pl_filter_sinc,        "Sinc (unwindowed)"},
+    {"ginseng",             &pl_filter_ginseng,     "Ginseng (Jinc-Sinc)"},
+    {"ewa_jinc",            &pl_filter_ewa_jinc,    "EWA Jinc (unwindowed)"},
+    {"ewa_ginseng",         &pl_filter_ewa_ginseng, "EWA Ginseng"},
+    {"ewa_hann",            &pl_filter_ewa_hann,    "EWA Hann"},
+    {"catmull_rom",         &pl_filter_catmull_rom, "Catmull-Rom"},
+    {"robidoux",            &pl_filter_robidoux,            "Robidoux"},
+    {"robidouxsharp",       &pl_filter_robidouxsharp,       "RobidouxSharp"},
+    {"ewa_robidoux",        &pl_filter_ewa_robidoux,        "EWA Robidoux"},
+    {"ewa_robidouxsharp",   &pl_filter_ewa_robidouxsharp,   "EWA RobidouxSharp"},
+
+    // Aliases
+    {"box",                 &pl_filter_nearest}, // alias
+    {"triangle",            &pl_filter_bilinear}, // alias
     {"ewa_hanning",         &pl_filter_ewa_hann}, // alias
-    {"haasnsoft",           &pl_filter_haasnsoft},
-    {"bicubic",             &pl_filter_bicubic},
-    {"catmull_rom",         &pl_filter_catmull_rom},
-    {"mitchell",            &pl_filter_mitchell},
-    {"robidoux",            &pl_filter_robidoux},
-    {"robidouxsharp",       &pl_filter_robidouxsharp},
-    {"ewa_robidoux",        &pl_filter_ewa_robidoux},
-    {"ewa_robidouxsharp",   &pl_filter_ewa_robidouxsharp},
     {0},
 };
+
+const int pl_num_filter_presets = PL_ARRAY_SIZE(pl_filter_presets) - 1;
