@@ -16,6 +16,7 @@
  */
 
 #include "common.h"
+#include "utils.h"
 #include "gpu.h"
 
 const struct pl_opengl_params pl_opengl_default_params = {0};
@@ -23,6 +24,7 @@ const struct pl_opengl_params pl_opengl_default_params = {0};
 struct priv {
     struct pl_context *ctx;
     bool is_debug;
+    bool is_debug_egl;
 };
 
 static void GLAPIENTRY debug_cb(GLenum source, GLenum type, GLuint id,
@@ -42,26 +44,55 @@ static void GLAPIENTRY debug_cb(GLenum source, GLenum type, GLuint id,
     pl_msg(ctx, level, "GL: %s", message);
 }
 
+#ifdef EPOXY_HAS_EGL
+
+static void debug_cb_egl(EGLenum error, const char *command,
+                         EGLint messageType, EGLLabelKHR threadLabel,
+                         EGLLabelKHR objectLabel, const char *message)
+{
+    struct pl_context *ctx = threadLabel;
+    if (!ctx)
+        return;
+
+    enum pl_log_level level = PL_LOG_ERR;
+    switch (messageType) {
+    case EGL_DEBUG_MSG_CRITICAL_KHR:    level = PL_LOG_FATAL; break;
+    case EGL_DEBUG_MSG_ERROR_KHR:       level = PL_LOG_ERR; break;
+    case EGL_DEBUG_MSG_WARN_KHR:        level = PL_LOG_WARN; break;
+    case EGL_DEBUG_MSG_INFO_KHR:        level = PL_LOG_DEBUG; break;
+    }
+
+    pl_msg(ctx, level, "EGL: %s: %s %s", command, egl_err_str(error),
+           message);
+}
+
+#endif // EPOXY_HAS_EGL
+
 void pl_opengl_destroy(const struct pl_opengl **ptr)
 {
     const struct pl_opengl *pl_gl = *ptr;
     if (!pl_gl)
         return;
 
-    struct priv *p = TA_PRIV(pl_gl);
+    struct priv *p = PL_PRIV(pl_gl);
     if (p->is_debug)
         glDebugMessageCallback(NULL, NULL);
 
+#ifdef EPOXY_HAS_EGL
+    if (p->is_debug_egl)
+        eglDebugMessageControlKHR(NULL, NULL);
+#endif
+
     pl_gpu_destroy(pl_gl->gpu);
-    TA_FREEP((void **) ptr);
+    pl_free_ptr((void **) ptr);
 }
 
 const struct pl_opengl *pl_opengl_create(struct pl_context *ctx,
                                          const struct pl_opengl_params *params)
 {
     params = PL_DEF(params, &pl_opengl_default_params);
-    struct pl_opengl *pl_gl = talloc_zero_priv(NULL, struct pl_opengl, struct priv);
-    struct priv *p = TA_PRIV(pl_gl);
+    struct pl_opengl *pl_gl = pl_zalloc_priv(NULL, struct pl_opengl, struct priv);
+    struct priv *p = PL_PRIV(pl_gl);
     p->ctx = ctx;
 
     int ver = epoxy_gl_version();
@@ -88,6 +119,19 @@ const struct pl_opengl *pl_opengl_create(struct pl_context *ctx,
         } else {
             PL_DEBUG(p, "    GL_EXTENSIONS: %s", glGetString(GL_EXTENSIONS));
         }
+
+#ifdef EPOXY_HAS_EGL
+        if (params->egl_display) {
+            PL_DEBUG(p, "    EGL_EXTENSIONS: %s",
+                     eglQueryString(params->egl_display, EGL_EXTENSIONS));
+        }
+#endif
+    }
+
+    if (!params->allow_software && gl_is_software()) {
+        PL_FATAL(p, "OpenGL context is suspected to be a software rasterizer, "
+                 "but `allow_software` is false.");
+        goto error;
     }
 
     if (params->debug) {
@@ -98,9 +142,27 @@ const struct pl_opengl *pl_opengl_create(struct pl_context *ctx,
             PL_WARN(p, "OpenGL debugging requested but GL_ARB_debug_output "
                     "unavailable.. ignoring!");
         }
+
+#ifdef EPOXY_HAS_EGL
+        if (params->egl_display && epoxy_has_egl_extension(params->egl_display, "EGL_KHR_debug")) {
+            static const EGLAttrib attribs[] = {
+                // Enable everything under the sun, because the `pl_ctx` log
+                // level may change at runtime.
+                EGL_DEBUG_MSG_CRITICAL_KHR, EGL_TRUE,
+                EGL_DEBUG_MSG_ERROR_KHR,    EGL_TRUE,
+                EGL_DEBUG_MSG_WARN_KHR,     EGL_TRUE,
+                EGL_DEBUG_MSG_INFO_KHR,     EGL_TRUE,
+                EGL_NONE,
+            };
+
+            eglDebugMessageControlKHR(debug_cb_egl, attribs);
+            eglLabelObjectKHR(NULL, EGL_OBJECT_THREAD_KHR, NULL, ctx);
+            p->is_debug_egl = true;
+        }
+#endif // EPOXY_HAS_EGL
     }
 
-    pl_gl->gpu = pl_gpu_create_gl(ctx);
+    pl_gl->gpu = pl_gpu_create_gl(ctx, params);
     if (!pl_gl->gpu)
         goto error;
 
