@@ -26,13 +26,15 @@ struct pl_shader *pl_shader_alloc(struct pl_context *ctx,
                                   const struct pl_shader_params *params)
 {
     pl_assert(ctx);
-    struct pl_shader *sh = talloc_ptrtype(ctx, sh);
+    struct pl_shader *sh = pl_alloc_ptr(ctx, sh);
     *sh = (struct pl_shader) {
         .ctx = ctx,
         .mutable = true,
-        .tmp = talloc_ref_new(ctx),
     };
 
+    // Ensure there's always at least one `tmp` object
+    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(ctx));
+
     if (params)
         sh->res.params = *params;
 
@@ -45,21 +47,26 @@ void pl_shader_free(struct pl_shader **psh)
     if (!sh)
         return;
 
-    talloc_ref_deref(&sh->tmp);
-    TA_FREEP(psh);
+    for (int i = 0; i < sh->tmp.num; i++)
+        pl_ref_deref(&sh->tmp.elem[i]);
+
+    pl_free_ptr(psh);
 }
 
 void pl_shader_reset(struct pl_shader *sh, const struct pl_shader_params *params)
 {
+    for (int i = 0; i < sh->tmp.num; i++)
+        pl_ref_deref(&sh->tmp.elem[i]);
+
     struct pl_shader new = {
         .ctx = sh->ctx,
-        .tmp = talloc_ref_new(sh->ctx),
         .mutable = true,
 
         // Preserve array allocations
-        .variables      = sh->variables,
-        .descriptors    = sh->descriptors,
-        .vertex_attribs = sh->vertex_attribs,
+        .tmp.elem   = sh->tmp.elem,
+        .vars.elem  = sh->vars.elem,
+        .descs.elem = sh->descs.elem,
+        .vas.elem   = sh->vas.elem,
     };
 
     if (params)
@@ -67,10 +74,10 @@ void pl_shader_reset(struct pl_shader *sh, const struct pl_shader_params *params
 
     // Preserve buffer allocations
     for (int i = 0; i < PL_ARRAY_SIZE(new.buffers); i++)
-        new.buffers[i] = (struct bstr) { .start = sh->buffers[i].start };
+        new.buffers[i] = (pl_str) { .buf = sh->buffers[i].buf };
 
-    talloc_ref_deref(&sh->tmp);
     *sh = new;
+    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(sh->ctx));
 }
 
 bool pl_shader_is_failed(const struct pl_shader *sh)
@@ -175,7 +182,7 @@ uint64_t pl_shader_signature(const struct pl_shader *sh)
 {
     uint64_t res = 0;
     for (int i = 0; i < PL_ARRAY_SIZE(sh->buffers); i++)
-        res ^= bstr_hash64(sh->buffers[i]);
+        res ^= pl_str_hash(sh->buffers[i]);
 
     // FIXME: also hash in the configuration of the descriptors/variables
 
@@ -184,37 +191,43 @@ uint64_t pl_shader_signature(const struct pl_shader *sh)
 
 ident_t sh_fresh(struct pl_shader *sh, const char *name)
 {
-    return talloc_asprintf(sh->tmp, "_%s_%d_%u", PL_DEF(name, "var"),
-                           sh->fresh++, SH_PARAMS(sh).id);
+    return pl_asprintf(SH_TMP(sh), "_%s_%d_%u", PL_DEF(name, "var"),
+                       sh->fresh++, SH_PARAMS(sh).id);
 }
 
 ident_t sh_var(struct pl_shader *sh, struct pl_shader_var sv)
 {
     sv.var.name = sh_fresh(sh, sv.var.name);
-    sv.data = talloc_memdup(sh->tmp, sv.data, pl_var_host_layout(0, &sv.var).size);
-    TARRAY_APPEND(sh, sh->variables, sh->res.num_variables, sv);
+    sv.data = pl_memdup(SH_TMP(sh), sv.data, pl_var_host_layout(0, &sv.var).size);
+    PL_ARRAY_APPEND(sh, sh->vars, sv);
     return (ident_t) sv.var.name;
 }
 
 ident_t sh_desc(struct pl_shader *sh, struct pl_shader_desc sd)
 {
-    // Skip re-attaching the same buffer desc twice
-    // FIXME: define aliases if the variable names differ
     switch (sd.desc.type) {
     case PL_DESC_BUF_UNIFORM:
     case PL_DESC_BUF_STORAGE:
     case PL_DESC_BUF_TEXEL_UNIFORM:
     case PL_DESC_BUF_TEXEL_STORAGE:
-        for (int i = 0; i < sh->res.num_descriptors; i++) {
-            if (sh->descriptors[i].object == sd.object)
-                return (ident_t) sh->descriptors[i].desc.name;
+        // Skip re-attaching the same buffer desc twice
+        // FIXME: define aliases if the variable names differ
+        for (int i = 0; i < sh->descs.num; i++) {
+            if (sh->descs.elem[i].binding.object == sd.binding.object)
+                return (ident_t) sh->descs.elem[i].desc.name;
         }
 
-    default: break;
+        size_t bsize = sizeof(sd.buffer_vars[0]) * sd.num_buffer_vars;
+        sd.buffer_vars = pl_memdup(SH_TMP(sh), sd.buffer_vars, bsize);
+        break;
+
+    default:
+        pl_assert(!sd.num_buffer_vars);
+        break;
     }
 
     sd.desc.name = sh_fresh(sh, sd.desc.name);
-    TARRAY_APPEND(sh, sh->descriptors, sh->res.num_descriptors, sd);
+    PL_ARRAY_APPEND(sh, sh->descs, sd);
     return (ident_t) sd.desc.name;
 }
 
@@ -240,7 +253,7 @@ ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
         { rc->x1, rc->y1 },
     };
 
-    float *data = talloc_memdup(sh->tmp, &vals[0][0], sizeof(vals));
+    float *data = pl_memdup(SH_TMP(sh), &vals[0][0], sizeof(vals));
     struct pl_shader_va va = {
         .attr = {
             .name     = sh_fresh(sh, name),
@@ -249,11 +262,13 @@ ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
         .data = { &data[0], &data[2], &data[4], &data[6] },
     };
 
-    TARRAY_APPEND(sh, sh->vertex_attribs, sh->res.num_vertex_attribs, va);
+    PL_ARRAY_APPEND(sh, sh->vas, va);
     return (ident_t) va.attr.name;
 }
 
 ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
+                enum pl_tex_address_mode address_mode,
+                enum pl_tex_sample_mode sample_mode,
                 const char *name, const struct pl_rect2df *rect,
                 ident_t *out_pos, ident_t *out_size, ident_t *out_pt)
 {
@@ -272,7 +287,11 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
             .name = name,
             .type = PL_DESC_SAMPLED_TEX,
         },
-        .object = tex,
+        .binding = {
+            .object = tex,
+            .address_mode = address_mode,
+            .sample_mode = sample_mode,
+        },
     });
 
     float sx, sy;
@@ -314,7 +333,7 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
     return itex;
 }
 
-bool sh_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
+bool sh_buf_desc_append(void *alloc, const struct pl_gpu *gpu,
                         struct pl_shader_desc *buf_desc,
                         struct pl_var_layout *out_layout,
                         const struct pl_var new_var)
@@ -338,7 +357,7 @@ bool sh_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
 
     if (out_layout)
         *out_layout = bv.layout;
-    TARRAY_APPEND(tactx, buf_desc->buffer_vars, buf_desc->num_buffer_vars, bv);
+    PL_ARRAY_APPEND_RAW(alloc, buf_desc->buffer_vars, buf_desc->num_buffer_vars, bv);
     return true;
 }
 
@@ -352,22 +371,20 @@ size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc)
     return last->layout.offset + last->layout.size;
 }
 
-void pl_shader_append(struct pl_shader *sh, enum pl_shader_buf buf,
-                      const char *fmt, ...)
+void sh_append(struct pl_shader *sh, enum pl_shader_buf buf, const char *fmt, ...)
 {
     pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
 
     va_list ap;
     va_start(ap, fmt);
-    bstr_xappend_vasprintf_c(sh, &sh->buffers[buf], fmt, ap);
+    pl_str_append_vasprintf_c(sh, &sh->buffers[buf], fmt, ap);
     va_end(ap);
 }
 
-void pl_shader_append_bstr(struct pl_shader *sh, enum pl_shader_buf buf,
-                           struct bstr str)
+void sh_append_str(struct pl_shader *sh, enum pl_shader_buf buf, pl_str str)
 {
     pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
-    bstr_xappend(sh, &sh->buffers[buf], str);
+    pl_str_append(sh, &sh->buffers[buf], str);
 }
 
 static const char *insigs[] = {
@@ -429,8 +446,8 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
     sh->output_h = res_h;
 
     // Append the prelude and header
-    bstr_xappend(sh, &sh->buffers[SH_BUF_PRELUDE], sub->buffers[SH_BUF_PRELUDE]);
-    bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER],  sub->buffers[SH_BUF_HEADER]);
+    pl_str_append(sh, &sh->buffers[SH_BUF_PRELUDE], sub->buffers[SH_BUF_PRELUDE]);
+    pl_str_append(sh, &sh->buffers[SH_BUF_HEADER],  sub->buffers[SH_BUF_HEADER]);
 
     // Append the body as a new header function
     ident_t name = sh_fresh(sh, "sub");
@@ -442,16 +459,15 @@ ident_t sh_subpass(struct pl_shader *sh, const struct pl_shader *sub)
     } else {
         GLSLH("%s %s(%s) {\n", outsigs[sub->res.output], name, insigs[sub->res.input]);
     }
-    bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_BODY]);
+    pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_BODY]);
     GLSLH("%s\n}\n\n", retvals[sub->res.output]);
 
     // Copy over all of the descriptors etc.
-    talloc_ref_attach(sh->tmp, sub->tmp);
-#define COPY(f) TARRAY_CONCAT(sh, sh->f, sh->res.num_##f, sub->f, sub->res.num_##f)
-    COPY(variables);
-    COPY(descriptors);
-    COPY(vertex_attribs);
-#undef COPY
+    for (int i = 0; i < sub->tmp.num; i++)
+        PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_dup(sub->tmp.elem[i]));
+    PL_ARRAY_CONCAT(sh, sh->vars, sub->vars);
+    PL_ARRAY_CONCAT(sh, sh->descs, sub->descs);
+    PL_ARRAY_CONCAT(sh, sh->vas, sub->vas);
 
     return name;
 }
@@ -473,15 +489,15 @@ static ident_t sh_split(struct pl_shader *sh)
     }
 
     if (sh->buffers[SH_BUF_BODY].len) {
-        bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_BODY]);
+        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_BODY]);
         sh->buffers[SH_BUF_BODY].len = 0;
-        sh->buffers[SH_BUF_BODY].start[0] = '\0'; // for sanity / efficiency
+        sh->buffers[SH_BUF_BODY].buf[0] = '\0'; // for sanity / efficiency
     }
 
     if (sh->buffers[SH_BUF_FOOTER].len) {
-        bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_FOOTER]);
+        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_FOOTER]);
         sh->buffers[SH_BUF_FOOTER].len = 0;
-        sh->buffers[SH_BUF_FOOTER].start[0] = '\0';
+        sh->buffers[SH_BUF_FOOTER].buf[0] = '\0';
     }
 
     GLSLH("%s\n}\n\n", retvals[sh->res.output]);
@@ -505,16 +521,19 @@ const struct pl_shader_res *pl_shader_finalize(struct pl_shader *sh)
     GLSLP("\n");
 
     // Concatenate the header onto the prelude to form the final output
-    struct bstr *glsl = &sh->buffers[SH_BUF_PRELUDE];
-    bstr_xappend(sh, glsl, sh->buffers[SH_BUF_HEADER]);
+    pl_str *glsl = &sh->buffers[SH_BUF_PRELUDE];
+    pl_str_append(sh, glsl, sh->buffers[SH_BUF_HEADER]);
 
     // Set the vas/vars/descs
-    sh->res.vertex_attribs = sh->vertex_attribs;
-    sh->res.variables = sh->variables;
-    sh->res.descriptors = sh->descriptors;
+    sh->res.vertex_attribs = sh->vas.elem;
+    sh->res.num_vertex_attribs = sh->vas.num;
+    sh->res.variables = sh->vars.elem;
+    sh->res.num_variables = sh->vars.num;
+    sh->res.descriptors = sh->descs.elem;
+    sh->res.num_descriptors = sh->descs.num;
 
     // Update the result pointer and return
-    sh->res.glsl = glsl->start;
+    sh->res.glsl = glsl->buf;
     sh->mutable = false;
     return &sh->res;
 }
@@ -574,7 +593,7 @@ void pl_shader_obj_destroy(struct pl_shader_obj **ptr)
         obj->uninit(obj->gpu, obj->priv);
 
     *ptr = NULL;
-    talloc_free(obj);
+    pl_free(obj);
 }
 
 void *sh_require_obj(struct pl_shader *sh, struct pl_shader_obj **ptr,
@@ -597,10 +616,10 @@ void *sh_require_obj(struct pl_shader *sh, struct pl_shader_obj **ptr,
     }
 
     if (!obj) {
-        obj = talloc_zero(NULL, struct pl_shader_obj);
+        obj = pl_zalloc_ptr(NULL, obj);
         obj->gpu = SH_GPU(sh);
         obj->type = type;
-        obj->priv = talloc_zero_size(obj, priv_size);
+        obj->priv = pl_zalloc(obj, priv_size);
         obj->uninit = uninit;
     }
 
@@ -668,30 +687,23 @@ static ident_t sh_lut_pos(struct pl_shader *sh, int lut_size)
 
 struct sh_lut_obj {
     enum sh_lut_method method;
+    enum pl_var_type type;
+    bool linear;
     int width, height, depth, comps;
-    union {
-        const struct pl_tex *tex;
-        struct bstr str;
-        float *data;
-    } weights;
+    uint64_t signature;
+
+    // weights, depending on the method
+    const struct pl_tex *tex;
+    pl_str str;
+    void *data;
 };
 
 static void sh_lut_uninit(const struct pl_gpu *gpu, void *ptr)
 {
     struct sh_lut_obj *lut = ptr;
-    switch (lut->method) {
-    case SH_LUT_TEXTURE:
-    case SH_LUT_LINEAR:
-        pl_tex_destroy(gpu, &lut->weights.tex);
-        break;
-    case SH_LUT_UNIFORM:
-        talloc_free(lut->weights.data);
-        break;
-    case SH_LUT_LITERAL:
-        talloc_free(lut->weights.str.start);
-        break;
-    default: break;
-    }
+    pl_tex_destroy(gpu, &lut->tex);
+    pl_free(lut->str.buf);
+    pl_free(lut->data);
 
     *lut = (struct sh_lut_obj) {0};
 }
@@ -699,19 +711,20 @@ static void sh_lut_uninit(const struct pl_gpu *gpu, void *ptr)
 // Maximum number of floats to embed as a literal array (when using SH_LUT_AUTO)
 #define SH_LUT_MAX_LITERAL 256
 
-ident_t sh_lut(struct pl_shader *sh, struct pl_shader_obj **obj,
-               enum sh_lut_method method, int width, int height, int depth,
-               int comps, bool update, bool dynamic, void *priv,
-               void (*fill)(void *priv, float *data, int w, int h, int d))
+ident_t sh_lut(struct pl_shader *sh, const struct sh_lut_params *params)
 {
     const struct pl_gpu *gpu = SH_GPU(sh);
-    float *tmp = NULL;
+    void *tmp = NULL;
     ident_t ret = NULL;
 
-    pl_assert(width > 0 && height >= 0 && depth >= 0);
-    int sizes[] = { width, height, depth };
-    int size = width * PL_DEF(height, 1) * PL_DEF(depth, 1);
-    int dims = depth ? 3 : height ? 2 : 1;
+    pl_assert(params->width > 0 && params->height >= 0 && params->depth >= 0);
+    pl_assert(params->comps > 0);
+    pl_assert(params->type);
+    pl_assert(!params->linear || params->type == PL_VAR_FLOAT);
+
+    int sizes[] = { params->width, params->height, params->depth };
+    int size = params->width * PL_DEF(params->height, 1) * PL_DEF(params->depth, 1);
+    int dims = params->depth ? 3 : params->height ? 2 : 1;
 
     int texdim = 0;
     uint32_t max_tex_dim[] = {
@@ -737,7 +750,25 @@ ident_t sh_lut(struct pl_shader *sh, struct pl_shader_obj **obj,
 next_dim: ; // `continue` out of the inner loop
     }
 
-    struct sh_lut_obj *lut = SH_OBJ(sh, obj, PL_SHADER_OBJ_LUT,
+    static const enum pl_fmt_type fmt_type[PL_VAR_TYPE_COUNT] = {
+        [PL_VAR_SINT]   = PL_FMT_SINT,
+        [PL_VAR_UINT]   = PL_FMT_UINT,
+        [PL_VAR_FLOAT]  = PL_FMT_FLOAT,
+    };
+
+    enum pl_fmt_caps texcaps = PL_FMT_CAP_SAMPLEABLE;
+    if (params->linear)
+        texcaps |= PL_FMT_CAP_LINEAR;
+
+    const struct pl_fmt *texfmt = NULL;
+    if (texdim) {
+        texfmt = pl_find_fmt(gpu, fmt_type[params->type], params->comps,
+                             params->type == PL_VAR_FLOAT ? 16 : 32,
+                             pl_var_type_size(params->type) * 8,
+                             texcaps);
+    }
+
+    struct sh_lut_obj *lut = SH_OBJ(sh, params->object, PL_SHADER_OBJ_LUT,
                                     struct sh_lut_obj, sh_lut_uninit);
 
     if (!lut) {
@@ -745,97 +776,85 @@ next_dim: ; // `continue` out of the inner loop
         goto error;
     }
 
-    if (!gpu && method == SH_LUT_LINEAR) {
-        SH_FAIL(sh, "Linear LUTs require the use of a GPU!");
-        goto error;
+    enum sh_lut_method method = params->method;
+
+    // The linear sampling code currently only supports 1D linear interpolation
+    if (params->linear && dims > 1) {
+        if (texfmt) {
+            method = SH_LUT_TEXTURE;
+        } else {
+            SH_FAIL(sh, "Can't emulate linear LUTs for 2D/3D LUTs and no "
+                    "texture support available!");
+            goto error;
+        }
     }
 
-    if (!gpu) {
-        PL_TRACE(sh, "No GPU available, falling back to literal LUT embedding");
-        method = SH_LUT_LITERAL;
-    }
+    // Older GLSL forbids literal array constructors
+    bool can_literal = sh_glsl(sh).version > 110;
 
     // Pick the best method
-    if (!method && size <= SH_LUT_MAX_LITERAL)
-        method = SH_LUT_LITERAL;
+    if (!method && size <= SH_LUT_MAX_LITERAL && !params->dynamic && can_literal)
+        method = SH_LUT_LITERAL; // use literals for small constant LUTs
 
-    if (!method && texdim)
-        method = SH_LUT_TEXTURE;
+    if (!method && texfmt)
+        method = SH_LUT_TEXTURE; // use textures if a texfmt exists
 
-    if (!method && gpu && gpu->caps & PL_GPU_CAP_INPUT_VARIABLES)
+    // Use an input variable as a last fallback
+    if (!method)
         method = SH_LUT_UNIFORM;
 
-    // texelFetch requires GLSL >= 130, so fall back to the LINEAR code
-    if (method == SH_LUT_TEXTURE && gpu->glsl.version < 130)
-        method = SH_LUT_LINEAR;
-
-    // No other method found
-    if (!method) {
-        PL_TRACE(sh, "No other LUT method works, falling back to literal "
-                 "embedding.. this is most likely a slow path!");
-        method = SH_LUT_LITERAL;
-    }
-
     // Forcibly reinitialize the existing LUT if needed
-    if (method != lut->method || width != lut->width || height != lut->height
-        || depth != lut->depth || comps != lut->comps)
+    bool update = params->update || lut->signature != params->signature;
+    if (method != lut->method || params->type != lut->type ||
+        params->linear != lut->linear || params->width != lut->width ||
+        params->height != lut->height || params->depth != lut->depth ||
+        params->comps != lut->comps)
     {
-        PL_DEBUG(sh, "LUT method or size changed, reinitializing..");
+        PL_DEBUG(sh, "LUT cache invalidated, regenerating..");
         update = true;
     }
 
     if (update) {
-        tmp = talloc_zero_size(NULL, size * comps * sizeof(float));
-        fill(priv, tmp, width, height, depth);
+        size_t buf_size = size * params->comps * pl_var_type_size(params->type);
+        tmp = pl_zalloc(NULL, buf_size);
+        params->fill(tmp, params);
 
         switch (method) {
-        case SH_LUT_TEXTURE:
-        case SH_LUT_LINEAR: {
+        case SH_LUT_TEXTURE: {
             if (!texdim) {
                 SH_FAIL(sh, "Texture LUT exceeds texture dimensions!");
                 goto error;
             }
 
-            enum pl_fmt_caps caps = PL_FMT_CAP_SAMPLEABLE;
-            enum pl_tex_sample_mode mode = PL_TEX_SAMPLE_NEAREST;
-
-            if (method == SH_LUT_LINEAR) {
-                caps |= PL_FMT_CAP_LINEAR;
-                mode = PL_TEX_SAMPLE_LINEAR;
-            }
-
-            const struct pl_fmt *fmt;
-            fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, comps, 16, 32, caps);
-            if (!fmt) {
+            if (!texfmt) {
                 SH_FAIL(sh, "Found no compatible texture format for LUT!");
                 goto error;
             }
 
-            struct pl_tex_params params = {
-                .w              = width,
-                .h              = PL_DEF(height, texdim >= 2 ? 1 : 0),
-                .d              = PL_DEF(depth,  texdim >= 3 ? 1 : 0),
-                .format         = fmt,
+            struct pl_tex_params tex_params = {
+                .w              = params->width,
+                .h              = PL_DEF(params->height, texdim >= 2 ? 1 : 0),
+                .d              = PL_DEF(params->depth,  texdim >= 3 ? 1 : 0),
+                .format         = texfmt,
                 .sampleable     = true,
-                .sample_mode    = mode,
-                .address_mode   = PL_TEX_ADDRESS_CLAMP,
-                .host_writable  = dynamic,
-                .initial_data   = dynamic ? NULL : tmp,
+                .host_writable  = params->dynamic,
+                .initial_data   = params->dynamic ? NULL : tmp,
             };
 
             bool ok;
-            if (dynamic) {
-                ok = pl_tex_recreate(gpu, &lut->weights.tex, &params);
+            if (params->dynamic) {
+                ok = pl_tex_recreate(gpu, &lut->tex, &tex_params);
                 if (ok) {
                     ok = pl_tex_upload(gpu, &(struct pl_tex_transfer_params) {
-                        .tex = lut->weights.tex,
+                        .tex = lut->tex,
                         .ptr = tmp,
                     });
                 }
             } else {
-                pl_tex_destroy(gpu, &lut->weights.tex);
-                lut->weights.tex = pl_tex_create(gpu, &params);
-                ok = lut->weights.tex;
+                // Can't use pl_tex_recreate because of `initial_data`
+                pl_tex_destroy(gpu, &lut->tex);
+                lut->tex = pl_tex_create(gpu, &tex_params);
+                ok = lut->tex;
             }
 
             if (!ok) {
@@ -846,25 +865,48 @@ next_dim: ; // `continue` out of the inner loop
         }
 
         case SH_LUT_UNIFORM:
-            talloc_free(lut->weights.data);
-            lut->weights.data = tmp; // re-use `tmp`
+            pl_free(lut->data);
+            lut->data = tmp; // re-use `tmp`
             tmp = NULL;
             break;
 
         case SH_LUT_LITERAL: {
-            lut->weights.str.len = 0;
-            for (int i = 0; i < size * comps; i += comps) {
+            lut->str.len = 0;
+            static const char prefix[PL_VAR_TYPE_COUNT] = {
+                [PL_VAR_SINT]   = 'i',
+                [PL_VAR_UINT]   = 'u',
+                [PL_VAR_FLOAT]  = ' ',
+            };
+
+            for (int i = 0; i < size * params->comps; i += params->comps) {
                 if (i > 0)
-                    bstr_xappend_asprintf_c(lut, &lut->weights.str, ",");
-                if (comps > 1)
-                    bstr_xappend_asprintf_c(lut, &lut->weights.str, "vec%d(", comps);
-                for (int c = 0; c < comps; c++) {
-                    bstr_xappend_asprintf_c(lut, &lut->weights.str, "%s%f",
-                                            c > 0 ? "," : "",
-                                            tmp[i+c]);
+                    pl_str_append_asprintf_c(lut, &lut->str, ",");
+                if (params->comps > 1) {
+                    pl_str_append_asprintf_c(lut, &lut->str, "%cvec%d(",
+                                             prefix[params->type], params->comps);
                 }
-                if (comps > 1)
-                    bstr_xappend_asprintf_c(lut, &lut->weights.str, ")");
+                for (int c = 0; c < params->comps; c++) {
+                    switch (params->type) {
+                    case PL_VAR_FLOAT:
+                        pl_str_append_asprintf_c(lut, &lut->str, "%s%f",
+                                                 c > 0 ? "," : "",
+                                                 ((float *) tmp)[i+c]);
+                        break;
+                    case PL_VAR_UINT:
+                        pl_str_append_asprintf_c(lut, &lut->str, "%s%u",
+                                                 c > 0 ? "," : "",
+                                                 ((unsigned int *) tmp)[i+c]);
+                        break;
+                    case PL_VAR_SINT:
+                        pl_str_append_asprintf_c(lut, &lut->str, "%s%d",
+                                                 c > 0 ? "," : "",
+                                                 ((int *) tmp)[i+c]);
+                        break;
+                    default: abort();
+                    }
+                }
+                if (params->comps > 1)
+                    pl_str_append_asprintf_c(lut, &lut->str, ")");
             }
             break;
         }
@@ -873,19 +915,24 @@ next_dim: ; // `continue` out of the inner loop
         }
 
         lut->method = method;
-        lut->width = width;
-        lut->height = height;
-        lut->depth = depth;
-        lut->comps = comps;
+        lut->type = params->type;
+        lut->linear = params->linear;
+        lut->width = params->width;
+        lut->height = params->height;
+        lut->depth = params->depth;
+        lut->comps = params->comps;
     }
 
     // Done updating, generate the GLSL
     ident_t name = sh_fresh(sh, "lut");
     ident_t arr_name = NULL;
 
-    static const char * const types[] = {"float", "vec2", "vec3", "vec4"};
-    static const char * const itypes[] = {"uint", "ivec2", "ivec3", "ivec4"};
     static const char * const swizzles[] = {"x", "xy", "xyz", "xyzw"};
+    static const char * const vartypes[PL_VAR_TYPE_COUNT][4] = {
+        [PL_VAR_SINT] = { "int", "ivec2", "ivec3", "ivec4" },
+        [PL_VAR_UINT] = { "uint", "uvec2", "uvec3", "uvec4" },
+        [PL_VAR_FLOAT] = { "float", "vec2", "vec3", "vec4" },
+    };
 
     switch (method) {
     case SH_LUT_TEXTURE: {
@@ -895,54 +942,48 @@ next_dim: ; // `continue` out of the inner loop
                 .name = "weights",
                 .type = PL_DESC_SAMPLED_TEX,
             },
-            .object = lut->weights.tex,
-        });
-
-        GLSLH("#define %s(pos) (texelFetch(%s, %s(pos",
-              name, tex, itypes[texdim - 1]);
-
-        // Fill up extra components of the index
-        for (int i = dims; i < texdim; i++)
-            GLSLH(", 0");
-
-        GLSLH("), 0).%s)\n", swizzles[comps - 1]);
-        ret = name;
-        break;
-    }
-
-    case SH_LUT_LINEAR: {
-        assert(texdim);
-        ident_t tex = sh_desc(sh, (struct pl_shader_desc) {
-            .desc = {
-                .name = "weights",
-                .type = PL_DESC_SAMPLED_TEX,
-            },
-            .object = lut->weights.tex,
-        });
-
-        ident_t pos_macros[PL_ARRAY_SIZE(sizes)] = {0};
-        for (int i = 0; i < dims; i++)
-            pos_macros[i] = sh_lut_pos(sh, sizes[i]);
-
-        GLSLH("#define %s(pos) (%s(%s, %s(\\\n",
-              name, sh_tex_fn(sh, lut->weights.tex->params),
-              tex, types[texdim - 1]);
-
-        for (int i = 0; i < texdim; i++) {
-            char sep = i == 0 ? ' ' : ',';
-            if (pos_macros[i]) {
-                if (dims > 1) {
-                    GLSLH("   %c%s(%s(pos).%c)\\\n", sep, pos_macros[i],
-                          types[dims - 1], "xyzw"[i]);
-                } else {
-                    GLSLH("   %c%s(float(pos))\\\n", sep, pos_macros[i]);
-                }
-            } else {
-                GLSLH("   %c%f\\\n", sep, 0.5);
+            .binding = {
+                .object = lut->tex,
+                .sample_mode = params->linear ? PL_TEX_SAMPLE_LINEAR
+                                              : PL_TEX_SAMPLE_NEAREST,
             }
+        });
+
+        // texelFetch requires GLSL >= 130, so fall back to the linear code
+        if (params->linear || gpu->glsl.version < 130) {
+            ident_t pos_macros[PL_ARRAY_SIZE(sizes)] = {0};
+            for (int i = 0; i < dims; i++)
+                pos_macros[i] = sh_lut_pos(sh, sizes[i]);
+
+            GLSLH("#define %s(pos) (%s(%s, %s(\\\n",
+                  name, sh_tex_fn(sh, lut->tex->params),
+                  tex, vartypes[PL_VAR_FLOAT][texdim - 1]);
+
+            for (int i = 0; i < texdim; i++) {
+                char sep = i == 0 ? ' ' : ',';
+                if (pos_macros[i]) {
+                    if (dims > 1) {
+                        GLSLH("   %c%s(%s(pos).%c)\\\n", sep, pos_macros[i],
+                              vartypes[PL_VAR_FLOAT][dims - 1], "xyzw"[i]);
+                    } else {
+                        GLSLH("   %c%s(float(pos))\\\n", sep, pos_macros[i]);
+                    }
+                } else {
+                    GLSLH("   %c%f\\\n", sep, 0.5);
+                }
+            }
+            GLSLH("  )).%s)\n", swizzles[params->comps - 1]);
+        } else {
+            GLSLH("#define %s(pos) (texelFetch(%s, %s(pos",
+                  name, tex, vartypes[PL_VAR_SINT][texdim - 1]);
+
+            // Fill up extra components of the index
+            for (int i = dims; i < texdim; i++)
+                GLSLH(", 0");
+
+            GLSLH("), 0).%s)\n", swizzles[params->comps - 1]);
         }
-        GLSLH("  )).%s)\n", swizzles[comps - 1]);
-        ret = name;
+
         break;
     }
 
@@ -950,19 +991,21 @@ next_dim: ; // `continue` out of the inner loop
         arr_name = sh_var(sh, (struct pl_shader_var) {
             .var = {
                 .name = "weights",
-                .type = PL_VAR_FLOAT,
-                .dim_v = comps,
+                .type = params->type,
+                .dim_v = params->comps,
                 .dim_m = 1,
                 .dim_a = size,
             },
-            .data = lut->weights.data,
+            .data = lut->data,
         });
         break;
 
     case SH_LUT_LITERAL:
         arr_name = sh_fresh(sh, "weights");
-        GLSLH("const %s %s[%d] = float[](\n  ", types[comps - 1], arr_name, size);
-        bstr_xappend(sh, &sh->buffers[SH_BUF_HEADER], lut->weights.str);
+        GLSLH("const %s %s[%d] = %s[](\n  ",
+              vartypes[params->type][params->comps - 1], arr_name, size,
+              vartypes[params->type][params->comps - 1]);
+        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], lut->str);
         GLSLH(");\n");
         break;
 
@@ -970,19 +1013,38 @@ next_dim: ; // `continue` out of the inner loop
     }
 
     if (arr_name) {
-        GLSLH("#define %s(pos) (%s[int((pos).x)\\\n", name, arr_name);
-        int shift = width;
+        GLSLH("#define %s(pos) (%s[int((pos)%s)\\\n",
+              name, arr_name, dims > 1 ? "[0]" : "");
+        int shift = params->width;
         for (int i = 1; i < dims; i++) {
             GLSLH("    + %d * int((pos)[%d])\\\n", shift, i);
             shift *= sizes[i];
         }
         GLSLH("  ])\n");
-        ret = name;
+
+        if (params->linear) {
+            pl_assert(dims == 1);
+            pl_assert(params->type == PL_VAR_FLOAT);
+            ident_t arr_lut = name;
+            name = sh_fresh(sh, "lut_lin");
+            GLSLH("%s %s(float fpos) {                              \n"
+                  "    fpos = clamp(fpos, 0.0, 1.0) * %d.0;         \n"
+                  "    float fbase = floor(fpos);                   \n"
+                  "    float fceil = ceil(fpos);                    \n"
+                  "    float fcoord = fpos - fbase;                 \n"
+                  "    return mix(%s(fbase), %s(fceil), fcoord);    \n"
+                  "}                                                \n",
+                  vartypes[PL_VAR_FLOAT][params->comps - 1], name,
+                  size - 1,
+                  arr_lut, arr_lut);
+        }
     }
 
+    pl_assert(name);
+    ret = name;
     // fall through
 error:
-    talloc_free(tmp);
+    pl_free(tmp);
     return ret;
 }
 
