@@ -39,12 +39,12 @@ enum pl_shader_buf {
 struct pl_shader {
     struct pl_context *ctx;
     struct pl_shader_res res; // for accumulating some of the fields
-    struct xta_ref *tmp; // only used for var/va/desc names and var/va data
+    PL_ARRAY(struct pl_ref *) tmp; // only used for var/va/desc names and data
     bool failed;
     bool mutable;
     int output_w;
     int output_h;
-    struct bstr buffers[SH_BUF_COUNT];
+    pl_str buffers[SH_BUF_COUNT];
     bool is_compute;
     bool flexible_work_groups;
     enum pl_sampler_type sampler_type;
@@ -52,14 +52,15 @@ struct pl_shader {
     int fresh;
 
     // mutable versions of the fields from pl_shader_res
-    struct pl_shader_va *vertex_attribs;
-    struct pl_shader_var *variables;
-    struct pl_shader_desc *descriptors;
+    PL_ARRAY(struct pl_shader_va) vas;
+    PL_ARRAY(struct pl_shader_var) vars;
+    PL_ARRAY(struct pl_shader_desc) descs;
 };
 
 // Helper functions for convenience
 #define SH_PARAMS(sh) ((sh)->res.params)
 #define SH_GPU(sh) (SH_PARAMS(sh).gpu)
+#define SH_TMP(sh) ((sh)->tmp.elem[0])
 
 // Returns the GLSL description, defaulting to desktop 130.
 struct pl_glsl_desc sh_glsl(const struct pl_shader *sh);
@@ -102,6 +103,8 @@ ident_t sh_attr_vec2(struct pl_shader *sh, const char *name,
 // Note that for e.g. compute shaders, the vec2 out_pos might be a macro that
 // expands to an expensive computation, and should be cached by the user.
 ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
+                enum pl_tex_address_mode address_mode,
+                enum pl_tex_sample_mode sample_mode,
                 const char *name, const struct pl_rect2df *rect,
                 ident_t *out_pos, ident_t *out_size, ident_t *out_pt);
 
@@ -110,7 +113,7 @@ ident_t sh_bind(struct pl_shader *sh, const struct pl_tex *tex,
 // variable could be successfully added (which may fail if you try exceeding
 // the size limits of the buffer type). If successful, the layout is stored
 // in *out_layout (may be NULL).
-bool sh_buf_desc_append(void *tactx, const struct pl_gpu *gpu,
+bool sh_buf_desc_append(void *alloc, const struct pl_gpu *gpu,
                         struct pl_shader_desc *buf_desc,
                         struct pl_var_layout *out_layout,
                         const struct pl_var new_var);
@@ -119,17 +122,15 @@ size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc);
 
 
 // Underlying function for appending text to a shader
-void pl_shader_append(struct pl_shader *sh, enum pl_shader_buf buf,
-                      const char *fmt, ...)
-    PRINTF_ATTRIBUTE(3, 4);
+void sh_append(struct pl_shader *sh, enum pl_shader_buf buf, const char *fmt, ...)
+    PL_PRINTF(3, 4);
 
-void pl_shader_append_bstr(struct pl_shader *sh, enum pl_shader_buf buf,
-                           struct bstr str);
+void sh_append_str(struct pl_shader *sh, enum pl_shader_buf buf, pl_str str);
 
-#define GLSLP(...) pl_shader_append(sh, SH_BUF_PRELUDE, __VA_ARGS__)
-#define GLSLH(...) pl_shader_append(sh, SH_BUF_HEADER, __VA_ARGS__)
-#define GLSL(...)  pl_shader_append(sh, SH_BUF_BODY, __VA_ARGS__)
-#define GLSLF(...) pl_shader_append(sh, SH_BUF_FOOTER, __VA_ARGS__)
+#define GLSLP(...) sh_append(sh, SH_BUF_PRELUDE, __VA_ARGS__)
+#define GLSLH(...) sh_append(sh, SH_BUF_HEADER, __VA_ARGS__)
+#define GLSL(...)  sh_append(sh, SH_BUF_BODY, __VA_ARGS__)
+#define GLSLF(...) sh_append(sh, SH_BUF_FOOTER, __VA_ARGS__)
 
 // Requires that the share is mutable, has an output signature compatible
 // with the given input signature, as well as an output size compatible with
@@ -143,7 +144,7 @@ enum pl_shader_obj_type {
     PL_SHADER_OBJ_PEAK_DETECT,
     PL_SHADER_OBJ_SAMPLER,
     PL_SHADER_OBJ_DITHER,
-    PL_SHADER_OBJ_3DLUT,
+    PL_SHADER_OBJ_ICC,
     PL_SHADER_OBJ_LUT,
     PL_SHADER_OBJ_AV1_GRAIN,
 };
@@ -177,32 +178,53 @@ enum sh_lut_method {
     SH_LUT_TEXTURE,  // upload as texture
     SH_LUT_UNIFORM,  // uniform array
     SH_LUT_LITERAL,  // constant / literal array in shader source (fallback)
-
-    // this is never picked by SH_DATA_AUTO
-    SH_LUT_LINEAR,   // upload as linearly-sampleable texture
 };
 
-// Makes a table of float vecs values available as a shader variable, using an
-// a given method (falling back if needed). The resulting identifier can be
-// sampled directly as %s(pos), where pos is a vector with the right number of
+struct sh_lut_params {
+    struct pl_shader_obj **object;
+
+    // Method and type of the LUT we intend to generate.
+    enum sh_lut_method method;
+    enum pl_var_type type;
+
+    // LUT dimensions. Unused dimensions may be left as 0.
+    int width;
+    int height;
+    int depth;
+    int comps;
+
+    // If true, LUT takes a vecN coordinate and linearly interpolates values,
+    // rather than taking an ivecN. Requires `type == PL_VAR_FLOAT`!
+    bool linear;
+
+    // If true, the LUT will always be regenerated, even if the dimensions have
+    // not changed.
+    bool update;
+
+    // Alternate way of triggering shader invalidations. If the signature
+    // does not match the LUT's signature, it will be regenerated.
+    uint64_t signature;
+
+    // If set to true, shader objects will be preserved and updated in-place
+    // rather than being treated as read-only.
+    bool dynamic;
+
+    // Will be called with a zero-initialized buffer whenever the data needs to
+    // be computed, which happens whenever the size is changed, the shader
+    // object is invalidated, or `update` is set to true.
+    //
+    // Note: Interpretation of `data` is according to `pl_var_type`.
+    void (*fill)(void *data, const struct sh_lut_params *params);
+    void *priv;
+};
+
+// Makes a table of values available as a shader variable, using an a given
+// method (falling back if needed). The resulting identifier can be sampled
+// directly as %s(pos), where pos is a vector with the right number of
 // dimensions. `pos` must be an integer vector within the bounds of the array,
 // unless the method is `SH_LUT_LINEAR`, in which case it's a float vector that
 // gets interpolated and clamped as needed. Returns NULL on error.
-//
-// This function also acts as `sh_require_obj`, and uses the `buf`, `tex`
-// and `text` fields of the resulting `obj`. (The other fields may be used by
-// the caller)
-//
-// The `fill` function will be called with a zero-initialized buffer whenever
-// the data needs to be computed, which happens whenever the size is changed,
-// the shader object is invalidated, or `update` is set to true.
-//
-// If `dynamic` is set to true, shader objects will be preserved and updated
-// in-place rather than being treated as read-only.
-ident_t sh_lut(struct pl_shader *sh, struct pl_shader_obj **obj,
-               enum sh_lut_method method, int width, int height, int depth,
-               int components, bool update, bool dynamic, void *priv,
-               void (*fill)(void *priv, float *data, int w, int h, int d));
+ident_t sh_lut(struct pl_shader *sh, const struct sh_lut_params *params);
 
 // Returns a GLSL-version appropriate "bvec"-like type. For GLSL 130+, this
 // returns bvecN. For GLSL 120, this returns vecN instead. The intended use of
