@@ -33,28 +33,66 @@ static inline struct pl_tex_params src_params(const struct pl_sample_src *src)
     return (struct pl_tex_params) {
         .w = src->tex_w,
         .h = src->tex_h,
-        .sample_mode = src->mode,
     };
 }
 
+enum filter {
+    NEAREST = PL_TEX_SAMPLE_NEAREST,
+    LINEAR  = PL_TEX_SAMPLE_LINEAR,
+    BEST,
+    FASTEST,
+};
+
 // Helper function to compute the src/dst sizes and upscaling ratios
 static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
                       ident_t *src_tex, ident_t *pos, ident_t *size, ident_t *pt,
-                      float *ratio_x, float *ratio_y, int *components,
-                      float *scale, bool resizeable, const char **fn)
+                      float *ratio_x, float *ratio_y, uint8_t *comp_mask,
+                      float *scale, bool resizeable, const char **fn,
+                      enum filter filter)
 {
     enum pl_shader_sig sig;
     float src_w, src_h;
+    enum pl_tex_sample_mode sample_mode;
     if (src->tex) {
+        const struct pl_fmt *fmt = src->tex->params.format;
+        bool can_linear = fmt->caps & PL_FMT_CAP_LINEAR;
         pl_assert(pl_tex_params_dimension(src->tex->params) == 2);
         sig = PL_SHADER_SIG_NONE;
         src_w = pl_rect_w(src->rect);
         src_h = pl_rect_h(src->rect);
+        switch (filter) {
+        case FASTEST:
+        case NEAREST:
+            sample_mode = PL_TEX_SAMPLE_NEAREST;
+            break;
+        case LINEAR:
+            if (!can_linear) {
+                SH_FAIL(sh, "Trying to use a shader that requires linear "
+                        "sampling with a texture whose format (%s) does not "
+                        "support PL_FMT_CAP_LINEAR", fmt->name);
+                return false;
+            }
+            sample_mode = PL_TEX_SAMPLE_LINEAR;
+            break;
+        case BEST:
+            sample_mode = can_linear ? PL_TEX_SAMPLE_LINEAR : PL_TEX_SAMPLE_NEAREST;
+            break;
+        }
     } else {
         pl_assert(src->tex_w && src->tex_h);
         sig = PL_SHADER_SIG_SAMPLER;
         src_w = src->sampled_w;
         src_h = src->sampled_h;
+        if (filter == BEST || filter == FASTEST) {
+            sample_mode = src->mode;
+        } else {
+            sample_mode = (enum pl_tex_sample_mode) filter;
+            if (sample_mode != src->mode) {
+                SH_FAIL(sh, "Trying to use a shader that requires a different "
+                        "filter mode than the external sampler.");
+                return false;
+            }
+        }
     }
 
     src_w = PL_DEF(src_w, src_params(src).w);
@@ -72,11 +110,20 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
     if (scale)
         *scale = PL_DEF(src->scale, 1.0);
 
-    if (components) {
-        int tex_comps = 4;
-        if (src->tex)
-            tex_comps = src->tex->params.format->num_components;
-        *components = PL_DEF(src->components, tex_comps);
+    if (comp_mask) {
+        uint8_t tex_mask = 0x0Fu;
+        if (src->tex) {
+            // Mask containing only the number of components in the texture
+            tex_mask = (1 << src->tex->params.format->num_components) - 1;
+        }
+
+        uint8_t src_mask = src->component_mask;
+        if (!src_mask)
+            src_mask = (1 << PL_DEF(src->components, 4)) - 1;
+
+        // Only actually sample components that are both requested and
+        // available in the texture being sampled
+        *comp_mask = tex_mask & src_mask;
     }
 
     if (resizeable)
@@ -95,7 +142,8 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
         if (fn)
             *fn = sh_tex_fn(sh, src->tex->params);
 
-        *src_tex = sh_bind(sh, src->tex, "src_tex", &rect, pos, size, pt);
+        *src_tex = sh_bind(sh, src->tex, src->address_mode, sample_mode,
+                           "src_tex", &rect, pos, size, pt);
     } else {
         if (size) {
             *size = sh_var(sh, (struct pl_shader_var) {
@@ -122,6 +170,7 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
 
         pl_assert(src->format);
         switch (src->format) {
+        case PL_FMT_UNKNOWN:
         case PL_FMT_FLOAT:
         case PL_FMT_UNORM:
         case PL_FMT_SNORM: sh->sampler_prefix = ' '; break;
@@ -140,15 +189,11 @@ static bool setup_src(struct pl_shader *sh, const struct pl_sample_src *src,
 void pl_shader_deband(struct pl_shader *sh, const struct pl_sample_src *src,
                       const struct pl_deband_params *params)
 {
-    if (src_params(src).sample_mode != PL_TEX_SAMPLE_LINEAR) {
-        SH_FAIL(sh, "Debanding requires sample_mode = PL_TEX_SAMPLE_LINEAR!");
-        return;
-    }
-
     float scale;
     ident_t tex, pos, pt;
     const char *fn;
-    if (!setup_src(sh, src, &tex, &pos, NULL, &pt, NULL, NULL, NULL, &scale, true, &fn))
+    if (!setup_src(sh, src, &tex, &pos, NULL, &pt, NULL, NULL, NULL, &scale,
+                   true, &fn, LINEAR))
         return;
 
     GLSL("vec4 color;\n");
@@ -211,7 +256,8 @@ bool pl_shader_sample_direct(struct pl_shader *sh, const struct pl_sample_src *s
     float scale;
     ident_t tex, pos;
     const char *fn;
-    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, &scale, true, &fn))
+    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, &scale,
+                   true, &fn, BEST))
         return false;
 
     GLSL("// pl_shader_sample_direct          \n"
@@ -220,6 +266,36 @@ bool pl_shader_sample_direct(struct pl_shader *sh, const struct pl_sample_src *s
     return true;
 }
 
+bool pl_shader_sample_nearest(struct pl_shader *sh, const struct pl_sample_src *src)
+{
+    float scale;
+    ident_t tex, pos;
+    const char *fn;
+    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, &scale,
+                   true, &fn, NEAREST))
+        return false;
+
+    GLSL("// pl_shader_sample_nearest         \n"
+         "vec4 color = vec4(%f) * %s(%s, %s); \n",
+         scale, fn, tex, pos);
+    return true;
+}
+
+bool pl_shader_sample_bilinear(struct pl_shader *sh, const struct pl_sample_src *src)
+{
+    float scale;
+    ident_t tex, pos;
+    const char *fn;
+    if (!setup_src(sh, src, &tex, &pos, NULL, NULL, NULL, NULL, NULL, &scale,
+                   true, &fn, LINEAR))
+        return false;
+
+    GLSL("// pl_shader_sample_bilinear        \n"
+         "vec4 color = vec4(%f) * %s(%s, %s); \n",
+         scale, fn, tex, pos);
+    return true;
+}
+
 static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char *s)
 {
     // Explanation of how bicubic scaling with only 4 texel fetches is done:
@@ -240,16 +316,11 @@ static void bicubic_calcweights(struct pl_shader *sh, const char *t, const char
 
 bool pl_shader_sample_bicubic(struct pl_shader *sh, const struct pl_sample_src *src)
 {
-    if (src_params(src).sample_mode != PL_TEX_SAMPLE_LINEAR) {
-        SH_FAIL(sh, "Trying to use fast bicubic sampling from a texture without "
-                "PL_TEX_SAMPLE_LINEAR");
-        return false;
-    }
-
     ident_t tex, pos, size, pt;
     float rx, ry, scale;
     const char *fn;
-    if (!setup_src(sh, src, &tex, &pos, &size, &pt, &rx, &ry, NULL, &scale, true, &fn))
+    if (!setup_src(sh, src, &tex, &pos, &size, &pt, &rx, &ry, NULL, &scale,
+                   true, &fn, LINEAR))
         return false;
 
     if (rx < 1 || ry < 1) {
@@ -309,7 +380,7 @@ static bool filter_compat(const struct pl_filter *filter, float inv_scale,
 // `in` is the given identifier, and `idx` must be defined by the caller
 static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
                          const char *fn, ident_t tex, ident_t lut, int x, int y,
-                         int comps, ident_t in)
+                         uint8_t comp_mask, ident_t in)
 {
     // Since we can't know the subpixel position in advance, assume a
     // worst case scenario
@@ -332,12 +403,19 @@ static void polar_sample(struct pl_shader *sh, const struct pl_filter *filter,
          lut, filter->radius);
 
     if (in) {
-        for (int n = 0; n < comps; n++)
-            GLSL("color[%d] += w * %s%d[idx];\n", n, in, n);
+        for (uint8_t comps = comp_mask; comps;) {
+            uint8_t c = __builtin_ctz(comps);
+            GLSL("color[%d] += w * %s%d[idx]; \n", c, in, c);
+            comps &= ~(1 << c);
+        }
     } else {
-        GLSL("in0 = %s(%s, base + pt * vec2(%d.0, %d.0)); \n"
-             "color += vec4(w) * in0;                     \n",
+        GLSL("c = %s(%s, base + pt * vec2(%d.0, %d.0)); \n",
              fn, tex, x, y);
+        for (uint8_t comps = comp_mask; comps;) {
+            uint8_t c = __builtin_ctz(comps);
+            GLSL("color[%d] += w * c[%d]; \n", c, c);
+            comps &= ~(1 << c);
+        }
     }
 
     if (maybe_skippable)
@@ -359,13 +437,13 @@ static void sh_sampler_uninit(const struct pl_gpu *gpu, void *ptr)
     *obj = (struct sh_sampler_obj) {0};
 }
 
-static void fill_polar_lut(void *priv, float *data, int w, int h, int d)
+static void fill_polar_lut(void *data, const struct sh_lut_params *params)
 {
-    const struct sh_sampler_obj *obj = priv;
+    const struct sh_sampler_obj *obj = params->priv;
     const struct pl_filter *filt = obj->filter;
 
-    pl_assert(w == filt->params.lut_entries);
-    memcpy(data, filt->weights, w * sizeof(float));
+    pl_assert(params->width == filt->params.lut_entries && params->comps == 1);
+    memcpy(data, filt->weights, params->width * sizeof(float));
 }
 
 bool pl_shader_sample_polar(struct pl_shader *sh,
@@ -405,11 +483,12 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         has_compute = false;
     }
 
-    int comps;
+    uint8_t comp_mask;
     float rx, ry, scale;
     ident_t src_tex, pos, size, pt;
     const char *fn;
-    if (!setup_src(sh, src, &src_tex, &pos, &size, &pt, &rx, &ry, &comps, &scale, false, &fn))
+    if (!setup_src(sh, src, &src_tex, &pos, &size, &pt, &rx, &ry, &comp_mask,
+                   &scale, false, &fn, FASTEST))
         return false;
 
     struct sh_sampler_obj *obj;
@@ -445,13 +524,6 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         }
     }
 
-    ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, lut_entries, 0, 0, 1,
-                         update, false, obj, fill_polar_lut);
-    if (!lut) {
-        SH_FAIL(sh, "Failed initializing polar LUT!");
-        return false;
-    }
-
     GLSL("// pl_shader_sample_polar                     \n"
          "vec4 color = vec4(0.0);                       \n"
          "{                                             \n"
@@ -478,8 +550,32 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
         ih = (int) ceil(bh / ry) + padding + 1;
 
     ident_t in = NULL;
-    int shmem_req = iw * ih * comps * sizeof(float);
-    if (has_compute && sh_try_compute(sh, bw, bh, false, shmem_req)) {
+    int num_comps = __builtin_popcount(comp_mask);
+    int shmem_req = iw * ih * num_comps * sizeof(float);
+    bool is_compute = has_compute && sh_try_compute(sh, bw, bh, false, shmem_req);
+
+    // For compute shaders, which read the input texels primarily from shmem,
+    // using a texture-based LUT is better. For the fragment shader fallback
+    // code, which is primarily texture bound, the extra cost of LUT
+    // interpolation is worth the reduction in texel fetches.
+    ident_t lut = sh_lut(sh, &(struct sh_lut_params) {
+        .object = &obj->lut,
+        .method = is_compute ? SH_LUT_TEXTURE : SH_LUT_AUTO,
+        .type = PL_VAR_FLOAT,
+        .width = lut_entries,
+        .comps = 1,
+        .linear = true,
+        .update = update,
+        .fill = fill_polar_lut,
+        .priv = obj,
+    });
+
+    if (!lut) {
+        SH_FAIL(sh, "Failed initializing polar LUT!");
+        return false;
+    }
+
+    if (is_compute) {
         // Compute shader kernel
         GLSL("vec2 wpos = %s_map(gl_WorkGroupID * gl_WorkGroupSize);        \n"
              "vec2 wbase = wpos - pt * fract(wpos * size - vec2(0.5));      \n"
@@ -493,27 +589,31 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
              ih, bh, iw, bw, fn, src_tex, offset, offset);
 
         in = sh_fresh(sh, "in");
-        for (int c = 0; c < comps; c++) {
+        for (uint8_t comps = comp_mask; comps;) {
+            uint8_t c = __builtin_ctz(comps);
             GLSLH("shared float %s%d[%d];   \n", in, c, ih * iw);
             GLSL("%s%d[%d * y + x] = c[%d]; \n", in, c, iw, c);
+            comps &= ~(1 << c);
         }
 
-        GLSL("}}                    \n"
-             "groupMemoryBarrier(); \n"
-             "barrier();            \n");
+        GLSL("}}                     \n"
+             "barrier();             \n");
 
         // Dispatch the actual samples
         for (int y = 1 - bound; y <= bound; y++) {
             for (int x = 1 - bound; x <= bound; x++) {
                 GLSL("idx = %d * rel.y + rel.x + %d;\n",
                      iw, iw * (y + offset) + x + offset);
-                polar_sample(sh, obj->filter, fn, src_tex, lut, x, y, comps, in);
+                polar_sample(sh, obj->filter, fn, src_tex, lut, x, y, comp_mask, in);
             }
         }
     } else {
         // Fragment shader sampling
-        for (int n = 0; n < comps; n++)
-            GLSL("vec4 in%d;\n", n);
+        for (uint8_t comps = comp_mask; comps;) {
+            uint8_t c = __builtin_ctz(comps);
+            GLSL("vec4 in%d;\n", c);
+            comps &= ~(1 << c);
+        }
 
         // Iterate over the LUT space in groups of 4 texels at a time, and
         // decide for each texel group whether to use gathering or direct
@@ -538,16 +638,18 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
                     for (int yy = y; yy <= bound && yy <= y + 1; yy++) {
                         for (int xx = x; xx <= bound && xx <= x + 1; xx++) {
                             polar_sample(sh, obj->filter, fn, src_tex, lut,
-                                         xx, yy, comps, NULL);
+                                         xx, yy, comp_mask, NULL);
                         }
                     }
                     continue; // next group of 4
                 }
 
                 // Gather the four surrounding texels simultaneously
-                for (int n = 0; n < comps; n++) {
+                for (uint8_t comps = comp_mask; comps;) {
+                    uint8_t c = __builtin_ctz(comps);
                     GLSL("in%d = textureGatherOffset(%s, base, "
-                         "ivec2(%d, %d), %d);\n", n, src_tex, x, y, n);
+                         "ivec2(%d, %d), %d);\n", c, src_tex, x, y, c);
+                    comps &= ~(1 << c);
                 }
 
                 // Mix in all of the points with their weights
@@ -561,7 +663,7 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
 
                     GLSL("idx = %d;\n", p);
                     polar_sample(sh, obj->filter, fn, src_tex, lut,
-                                 x+xo[p], y+yo[p], comps, "in");
+                                 x+xo[p], y+yo[p], comp_mask, "in");
                 }
             }
         }
@@ -573,13 +675,14 @@ bool pl_shader_sample_polar(struct pl_shader *sh,
     return true;
 }
 
-static void fill_ortho_lut(void *priv, float *data, int w, int h, int d)
+static void fill_ortho_lut(void *data, const struct sh_lut_params *params)
 {
-    const struct sh_sampler_obj *obj = priv;
+    const struct sh_sampler_obj *obj = params->priv;
     const struct pl_filter *filt = obj->filter;
+    size_t entries = filt->params.lut_entries * filt->row_stride;
 
-    pl_assert(w * h * 4 == filt->params.lut_entries * filt->row_stride);
-    memcpy(data, filt->weights, w * h * 4 * sizeof(float));
+    pl_assert(params->width * params->height * params->comps == entries);
+    memcpy(data, filt->weights, entries * sizeof(float));
 }
 
 bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
@@ -610,15 +713,14 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
         abort();
     }
 
-    int comps;
-    float ratio[2], scale;
+    uint8_t comp_mask;
+    float ratio[PL_SEP_PASSES], scale;
     ident_t src_tex, pos, size, pt;
     const char *fn;
-    if (!setup_src(sh, &srcfix, &src_tex, &pos, &size, &pt, &ratio[1], &ratio[0],
-                   &comps, &scale, false, &fn))
-    {
+    if (!setup_src(sh, &srcfix, &src_tex, &pos, &size, &pt,
+                   &ratio[PL_SEP_HORIZ], &ratio[PL_SEP_VERT],
+                   &comp_mask, &scale, false, &fn, FASTEST))
         return false;
-    }
 
     // We can store a separate sampler object per dimension, so dispatch the
     // right one. This is needed for two reasons:
@@ -669,8 +771,17 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
 
     int N = obj->filter->row_size; // number of samples to convolve
     int width = obj->filter->row_stride / 4; // width of the LUT texture
-    ident_t lut = sh_lut(sh, &obj->lut, SH_LUT_LINEAR, width, lut_entries, 0, 4,
-                         update, false, obj, fill_ortho_lut);
+    ident_t lut = sh_lut(sh, &(struct sh_lut_params) {
+        .object = &obj->lut,
+        .type = PL_VAR_FLOAT,
+        .width = width,
+        .height = lut_entries,
+        .comps = 4,
+        .linear = true,
+        .update = update,
+        .fill = fill_ortho_lut,
+        .priv = obj,
+    });
     if (!lut) {
         SH_FAIL(sh, "Failed initializing separated LUT!");
         return false;
@@ -714,13 +825,19 @@ bool pl_shader_sample_ortho(struct pl_shader *sh, int pass,
         GLSL("weight = ws[%d];\n", n % 4);
 
         // Load the input texel and add it to the running sum
-        GLSL("c = %s(%s, base + pt * vec2(%d.0)); \n"
-             "color += vec4(weight) * c;          \n",
+        GLSL("c = %s(%s, base + pt * vec2(%d.0)); \n",
              fn, src_tex, n);
 
-        if (use_ar && (n == N / 2 - 1 || n == N / 2)) {
-            GLSL("lo = min(lo, c); \n"
-                 "hi = max(hi, c); \n");
+        for (uint8_t comps = comp_mask; comps;) {
+            uint8_t c = __builtin_ctz(comps);
+            GLSL("color[%d] += weight * c[%d]; \n", c, c);
+            comps &= ~(1 << c);
+
+            if (use_ar && (n == N / 2 - 1 || n == N / 2)) {
+                GLSL("lo[%d] = min(lo[%d], c[%d]); \n"
+                     "hi[%d] = max(hi[%d], c[%d]); \n",
+                     c, c, c, c, c, c);
+            }
         }
     }
 
