@@ -18,12 +18,12 @@ int main()
 
     struct pl_color_repr tv_repr = {
         .sys       = PL_COLOR_SYSTEM_BT_709,
-        .levels    = PL_COLOR_LEVELS_TV,
+        .levels    = PL_COLOR_LEVELS_LIMITED,
     };
 
     struct pl_color_repr pc_repr = {
         .sys       = PL_COLOR_SYSTEM_RGB,
-        .levels    = PL_COLOR_LEVELS_PC,
+        .levels    = PL_COLOR_LEVELS_FULL,
     };
 
     // Ensure this is a no-op for bits == bits
@@ -122,7 +122,7 @@ int main()
 
         printf("testing color system %u\n", (unsigned) sys);
         struct pl_color_repr repr = {
-            .levels = PL_COLOR_LEVELS_TV,
+            .levels = PL_COLOR_LEVELS_LIMITED,
             .sys = sys,
         };
 
@@ -304,4 +304,10 @@ int main()
 
     // These models should round-trip green
     TEST_CONE(pl_vision_normal, green);
+
+    // Color adaptation tests
+    struct pl_cie_xy d65 = pl_white_from_temp(6504);
+    struct pl_cie_xy d55 = pl_white_from_temp(5503);
+    REQUIRE(feq(d65.x, 0.31271, 1e-3) && feq(d65.y, 0.32902, 1e-3));
+    REQUIRE(feq(d55.x, 0.33242, 1e-3) && feq(d55.y, 0.34743, 1e-3));
 }
