@@ -1,25 +1,23 @@
 #include "tests.h"
 #include "shaders.h"
 
-static uint8_t test_src[16*16*16 * 4 * sizeof(double)] = {0};
-static uint8_t test_dst[16*16*16 * 4 * sizeof(double)] = {0};
-
 static void pl_buffer_tests(const struct pl_gpu *gpu)
 {
     const size_t buf_size = 1024;
-    assert(buf_size <= sizeof(test_src));
+    if (buf_size > gpu->limits.max_buf_size)
+        return;
 
+    uint8_t *test_src = malloc(buf_size * 2);
+    uint8_t *test_dst = test_src + buf_size;
+    assert(test_src && test_dst);
     memset(test_dst, 0, buf_size);
     for (int i = 0; i < buf_size; i++)
         test_src[i] = (RANDOM * 256);
 
-    const struct pl_buf *buf = NULL;
-    if (buf_size > gpu->limits.max_xfer_size)
-        return;
+    const struct pl_buf *buf = NULL, *tbuf = NULL;
 
-    printf("test xfer buffer static creation and readback\n");
+    printf("test buffer static creation and readback\n");
     buf = pl_buf_create(gpu, &(struct pl_buf_params) {
-        .type = PL_BUF_TEX_TRANSFER,
         .size = buf_size,
         .host_readable = true,
         .initial_data = test_src,
@@ -30,10 +28,9 @@ static void pl_buffer_tests(const struct pl_gpu *gpu)
     REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
     pl_buf_destroy(gpu, &buf);
 
-    printf("test xfer buffer empty creation, update and readback\n");
+    printf("test buffer empty creation, update and readback\n");
     memset(test_dst, 0, buf_size);
     buf = pl_buf_create(gpu, &(struct pl_buf_params) {
-        .type = PL_BUF_TEX_TRANSFER,
         .size = buf_size,
         .host_writable = true,
         .host_readable = true,
@@ -45,10 +42,28 @@ static void pl_buffer_tests(const struct pl_gpu *gpu)
     REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
     pl_buf_destroy(gpu, &buf);
 
+    printf("test buffer-buffer copy and readback\n");
+    memset(test_dst, 0, buf_size);
+    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .size = buf_size,
+        .initial_data = test_src,
+    });
+
+    tbuf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .size = buf_size,
+        .host_readable = true,
+    });
+
+    REQUIRE(buf && tbuf);
+    pl_buf_copy(gpu, tbuf, 0, buf, 0, buf_size);
+    REQUIRE(pl_buf_read(gpu, tbuf, 0, test_dst, buf_size));
+    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    pl_buf_destroy(gpu, &buf);
+    pl_buf_destroy(gpu, &tbuf);
+
     if (gpu->caps & PL_GPU_CAP_MAPPED_BUFFERS) {
         printf("test host mapped buffer readback\n");
         buf = pl_buf_create(gpu, &(struct pl_buf_params) {
-            .type = PL_BUF_TEX_TRANSFER,
             .size = buf_size,
             .host_mapped = true,
             .initial_data = test_src,
@@ -59,6 +74,14 @@ static void pl_buffer_tests(const struct pl_gpu *gpu)
         REQUIRE(memcmp(test_src, buf->data, buf_size) == 0);
         pl_buf_destroy(gpu, &buf);
     }
+
+    free(test_src);
+}
+
+static void test_cb(void *priv)
+{
+    bool *flag = priv;
+    *flag = true;
 }
 
 static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex[2],
@@ -85,26 +108,24 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     ul = pl_timer_create(gpu);
     dl = pl_timer_create(gpu);
 
+    bool ran_ul = false, ran_dl = false;
+
     REQUIRE(pl_tex_upload(gpu, &(struct pl_tex_transfer_params){
         .tex = tex[0],
         .ptr = src,
         .timer = ul,
+        .callback = (gpu->caps & PL_GPU_CAP_CALLBACKS) ? test_cb : NULL,
+        .priv = &ran_ul,
     }));
 
     // Test blitting, if possible for this format
     const struct pl_tex *dst_tex = tex[0];
     if (tex[0]->params.blit_src && tex[1]->params.blit_dst) {
-        struct pl_rect3d rc = {
-            .x0 = 0,
-            .y0 = 0,
-            .z0 = 0,
-            .x1 = tex[0]->params.w,
-            .y1 = tex[0]->params.h,
-            .z1 = tex[0]->params.d,
-        };
-
         pl_tex_clear(gpu, tex[1], (float[4]){0.0}); // for testing
-        pl_tex_blit(gpu, tex[1], tex[0], rc, rc);
+        pl_tex_blit(gpu, &(struct pl_tex_blit_params) {
+            .src = tex[0],
+            .dst = tex[1],
+        });
         dst_tex = tex[1];
     }
 
@@ -112,8 +133,14 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
         .tex = dst_tex,
         .ptr = dst,
         .timer = dl,
+        .callback = (gpu->caps & PL_GPU_CAP_CALLBACKS) ? test_cb : NULL,
+        .priv = &ran_dl,
     }));
 
+    pl_gpu_finish(gpu);
+    if (gpu->caps & PL_GPU_CAP_CALLBACKS)
+        REQUIRE(ran_ul && ran_dl);
+
     if (fmt->emulated && fmt->type == PL_FMT_FLOAT) {
         // TODO: can't memcmp here because bits might be lost due to the
         // emulated 16/32 bit upload paths, figure out a better way to
@@ -123,7 +150,6 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
     }
 
     // Report timer results
-    pl_gpu_finish(gpu);
     printf("upload time: %"PRIu64", download time: %"PRIu64"\n",
            pl_timer_query(gpu, ul), pl_timer_query(gpu, dl));
 
@@ -133,6 +159,10 @@ static void pl_test_roundtrip(const struct pl_gpu *gpu, const struct pl_tex *tex
 
 static void pl_texture_tests(const struct pl_gpu *gpu)
 {
+    const size_t max_size = 16*16*16 * 4 *sizeof(double);
+    uint8_t *test_src = malloc(max_size * 2);
+    uint8_t *test_dst = test_src + max_size;
+
     for (int f = 0; f < gpu->num_formats; f++) {
         const struct pl_fmt *fmt = gpu->formats[f];
         if (fmt->opaque || !(fmt->caps & PL_FMT_CAP_HOST_READABLE))
@@ -188,6 +218,8 @@ static void pl_texture_tests(const struct pl_gpu *gpu)
                 pl_tex_destroy(gpu, &tex[i]);
         }
     }
+
+    free(test_src);
 }
 
 static void pl_shader_tests(const struct pl_gpu *gpu)
@@ -242,7 +274,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     vert_fmt = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 3);
     REQUIRE(vert_fmt);
 
-    struct vertex { float pos[2]; float color[3]; } vertices[] = {
+    static const struct vertex { float pos[2]; float color[3]; } vertices[] = {
         {{-1.0, -1.0}, {0, 0, 0}},
         {{ 1.0, -1.0}, {1, 0, 0}},
         {{-1.0,  1.0}, {0, 1, 0}},
@@ -278,13 +310,11 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     pl_pass_run(gpu, &(struct pl_pass_run_params) {
         .pass           = pass,
         .target         = fbo,
+        .vertex_count   = PL_ARRAY_SIZE(vertices),
         .vertex_data    = vertices,
-        .vertex_count   = sizeof(vertices) / sizeof(struct vertex),
         .timer          = timer,
     });
 
-    pl_pass_destroy(gpu, &pass);
-
     // Wait until this pass is complete and report the timer result
     pl_gpu_finish(gpu);
     printf("timer query result: %"PRIu64"\n", pl_timer_query(gpu, timer));
@@ -314,9 +344,72 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     TEST_FBO_PATTERN(1e-6, "%s", "initial rendering");
 
+    if (sizeof(vertices) <= gpu->limits.max_vbo_size) {
+        // Test the use of an explicit vertex buffer
+        const struct pl_buf *vert = pl_buf_create(gpu, &(struct pl_buf_params) {
+            .size = sizeof(vertices),
+            .initial_data = vertices,
+            .drawable = true,
+        });
+
+        REQUIRE(vert);
+        pl_pass_run(gpu, &(struct pl_pass_run_params) {
+            .pass           = pass,
+            .target         = fbo,
+            .vertex_count   = sizeof(vertices) / sizeof(struct vertex),
+            .vertex_buf     = vert,
+            .buf_offset     = 0,
+        });
+
+        pl_buf_destroy(gpu, &vert);
+        TEST_FBO_PATTERN(1e-6, "%s", "using vertex buffer");
+    }
+
+    // Test the use of index buffers
+    static const uint16_t indices[] = { 3, 2, 1, 0 };
+    pl_pass_run(gpu, &(struct pl_pass_run_params) {
+        .pass           = pass,
+        .target         = fbo,
+        .vertex_count   = PL_ARRAY_SIZE(indices),
+        .vertex_data    = vertices,
+        .index_data     = indices,
+    });
+
+    pl_pass_destroy(gpu, &pass);
+    TEST_FBO_PATTERN(1e-6, "%s", "using indexed rendering");
+
     // Test the use of pl_dispatch
     struct pl_dispatch *dp = pl_dispatch_create(gpu->ctx, gpu);
-    struct pl_shader *sh;
+    struct pl_shader *sh = pl_dispatch_begin(dp);
+    REQUIRE(pl_shader_custom(sh, &(struct pl_custom_shader) {
+        .body       = "color = vec4(col, 1.0);",
+        .input      = PL_SHADER_SIG_NONE,
+        .output     = PL_SHADER_SIG_COLOR,
+    }));
+
+    REQUIRE(pl_dispatch_vertex(dp, &(struct pl_dispatch_vertex_params) {
+        .shader         = &sh,
+        .target         = fbo,
+        .vertex_stride  = sizeof(struct vertex),
+        .vertex_position_idx = 0,
+        .num_vertex_attribs = 2,
+        .vertex_attribs = (struct pl_vertex_attrib[]) {{
+            .name   = "pos",
+            .fmt    = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2),
+            .offset = offsetof(struct vertex, pos),
+        }, {
+            .name   = "col",
+            .fmt    = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 3),
+            .offset = offsetof(struct vertex, color),
+        }},
+
+        .vertex_type    = PL_PRIM_TRIANGLE_STRIP,
+        .vertex_coords  = PL_COORDS_NORMALIZED,
+        .vertex_count   = PL_ARRAY_SIZE(vertices),
+        .vertex_data    = vertices,
+    }));
+
+    TEST_FBO_PATTERN(1e-6, "%s", "using custom vertices");
 
     const struct pl_tex *src;
     src = pl_tex_create(gpu, &(struct pl_tex_params) {
@@ -324,14 +417,13 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         .w              = FBO_W,
         .h              = FBO_H,
         .sampleable     = true,
-        .sample_mode    = PL_TEX_SAMPLE_LINEAR,
         .initial_data   = data,
     });
 
     // Test encoding/decoding of all gamma functions, color spaces, etc.
     for (enum pl_color_transfer trc = 0; trc < PL_COLOR_TRC_COUNT; trc++) {
         sh = pl_dispatch_begin(dp);
-        pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
         pl_shader_delinearize(sh, trc);
         pl_shader_linearize(sh, trc);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
@@ -345,7 +437,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     for (enum pl_color_system sys = 0; sys < PL_COLOR_SYSTEM_COUNT; sys++) {
         sh = pl_dispatch_begin(dp);
-        pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
         pl_shader_encode_color(sh, &(struct pl_color_repr) { .sys = sys });
         pl_shader_decode_color(sh, &(struct pl_color_repr) { .sys = sys }, NULL);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
@@ -377,7 +469,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         sh = pl_dispatch_begin(dp);
         struct pl_color_space src_space = { .light = light };
         struct pl_color_space dst_space = { 0 };
-        pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+        pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
         pl_shader_color_map(sh, NULL, src_space, dst_space, NULL, false);
         pl_shader_color_map(sh, NULL, dst_space, src_space, NULL, false);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
@@ -390,6 +482,30 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
 
     // Repeat this a few times to test the caching
     for (int i = 0; i < 10; i++) {
+        if (i == 5) {
+            printf("Recreating pl_dispatch to test the caching\n");
+            size_t size = pl_dispatch_save(dp, NULL);
+            REQUIRE(size > 0);
+            uint8_t *cache = malloc(size);
+            REQUIRE(cache);
+            REQUIRE(pl_dispatch_save(dp, cache) == size);
+
+            pl_dispatch_destroy(&dp);
+            dp = pl_dispatch_create(gpu->ctx, gpu);
+            pl_dispatch_load(dp, cache);
+
+#ifndef MSAN
+            // Test to make sure the pass regenerates the same cache, but skip
+            // this on MSAN because it doesn't like it when we read from
+            // program cache data generated by the non-instrumented GPU driver
+            uint64_t hash = pl_str_hash((pl_str) { cache, size });
+            REQUIRE(pl_dispatch_save(dp, NULL) == size);
+            REQUIRE(pl_dispatch_save(dp, cache) == size);
+            REQUIRE(pl_str_hash((pl_str) { cache, size }) == hash);
+#endif
+            free(cache);
+        }
+
         sh = pl_dispatch_begin(dp);
 
         // For testing, force the use of CS if possible
@@ -415,18 +531,55 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
         TEST_FBO_PATTERN(1e-6, "deband iter %d", i);
     }
 
-#ifdef PL_HAVE_LCMS
-    // Test the use of 3DLUTs if available
+    // Test peak detection and readback if possible
     sh = pl_dispatch_begin(dp);
-    pl_shader_sample_direct(sh, &(struct pl_sample_src) { .tex = src });
+    pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
 
-    struct pl_shader_obj *lut3d = NULL;
-    struct pl_3dlut_profile src_color = { .color = pl_color_space_bt709 };
-    struct pl_3dlut_profile dst_color = { .color = pl_color_space_srgb };
-    struct pl_3dlut_result out;
+    struct pl_shader_obj *peak_state = NULL;
+    if (pl_shader_detect_peak(sh, pl_color_space_monitor, &peak_state, NULL)) {
+        REQUIRE(pl_dispatch_compute(dp, &(struct pl_dispatch_compute_params) {
+            .shader = &sh,
+            .width = fbo->params.w,
+            .height = fbo->params.h,
+        }));
 
-    if (pl_3dlut_update(sh, &src_color, &dst_color, &lut3d, &out, NULL)) {
-        pl_3dlut_apply(sh, &lut3d);
+        float peak, avg;
+        REQUIRE(pl_get_detected_peak(peak_state, &peak, &avg));
+        printf("detected peak: %f, average: %f\n", peak, avg);
+
+        float real_peak = 0, real_avg = 0;
+        for (int y = 0; y < FBO_H; y++) {
+            for (int x = 0; x < FBO_W; x++) {
+                float *color = &data[(y * FBO_W + x) * 4];
+                float smax = powf(PL_MAX(color[0], color[1]), 2.2);
+                float slog = logf(PL_MAX(smax, 0.001));
+                real_peak = PL_MAX(smax, real_peak);
+                real_avg += slog;
+            }
+        }
+
+        real_peak *= 1.0 + pl_peak_detect_default_params.overshoot_margin;
+        real_avg = expf(real_avg / (FBO_W * FBO_H));
+        printf("real peak: %f, real average: %f\n", real_peak, real_avg);
+        REQUIRE(feq(peak, real_peak, 1e-4));
+        REQUIRE(feq(avg, real_avg, 1e-3));
+    }
+
+    pl_dispatch_abort(dp, &sh);
+    pl_shader_obj_destroy(&peak_state);
+
+#ifdef PL_HAVE_LCMS
+    // Test the use of ICC profiles if available
+    sh = pl_dispatch_begin(dp);
+    pl_shader_sample_nearest(sh, &(struct pl_sample_src) { .tex = src });
+
+    struct pl_shader_obj *icc = NULL;
+    struct pl_icc_color_space src_color = { .color = pl_color_space_bt709 };
+    struct pl_icc_color_space dst_color = { .color = pl_color_space_srgb };
+    struct pl_icc_result out;
+
+    if (pl_icc_update(sh, &src_color, &dst_color, &icc, &out, NULL)) {
+        pl_icc_apply(sh, &icc);
         REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
             .shader = &sh,
             .target = fbo,
@@ -434,7 +587,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     }
 
     pl_dispatch_abort(dp, &sh);
-    pl_shader_obj_destroy(&lut3d);
+    pl_shader_obj_destroy(&icc);
 #endif
 
     // Test AV1 grain synthesis
@@ -447,7 +600,7 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
             .component_mapping = { 0, 1, 2 },
             .repr = &(struct pl_color_repr) {
                 .sys = PL_COLOR_SYSTEM_BT_709,
-                .levels = PL_COLOR_LEVELS_TV,
+                .levels = PL_COLOR_LEVELS_LIMITED,
                 .bits = { .color_depth = 10, .sample_depth = 10 },
             },
         };
@@ -463,6 +616,35 @@ static void pl_shader_tests(const struct pl_gpu *gpu)
     }
     pl_shader_obj_destroy(&grain);
 
+    // Test custom shaders
+    struct pl_custom_shader custom = {
+        .header =
+            "vec3 invert(vec3 color)            \n"
+            "{                                  \n"
+            "    return vec3(1.0) - color;      \n"
+            "}                                  \n",
+
+        .body =
+            "color = vec4(gl_FragCoord.xy, 0.0, 1.0);   \n"
+            "color.rgb = invert(color.rgb) + offset;    \n",
+
+        .input = PL_SHADER_SIG_NONE,
+        .output = PL_SHADER_SIG_COLOR,
+
+        .num_variables = 1,
+        .variables = &(struct pl_shader_var) {
+            .var = pl_var_float("offset"),
+            .data = &(float) { 0.1 },
+        },
+    };
+
+    sh = pl_dispatch_begin(dp);
+    REQUIRE(pl_shader_custom(sh, &custom));
+    REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
+        .shader = &sh,
+        .target = fbo,
+    }));
+
     pl_dispatch_destroy(&dp);
     pl_tex_destroy(gpu, &src);
     pl_tex_destroy(gpu, &fbo);
@@ -494,8 +676,6 @@ static void pl_scaler_tests(const struct pl_gpu *gpu)
         .h              = 5,
         .format         = src_fmt,
         .sampleable     = true,
-        .sample_mode    = PL_TEX_SAMPLE_LINEAR,
-        .address_mode   = PL_TEX_ADDRESS_CLAMP,
         .initial_data   = &data_5x5[0][0],
     });
 
@@ -608,17 +788,84 @@ static const char *user_shader_tests[] = {
     "//!FORMAT rgba32f                                                      \n"
     "//!FILTER NEAREST                                                      \n"
     "//!BORDER REPEAT                                                       \n"
-    "0000803f000000000000000000000000000000000000803f000000000000000000000000000000000000803f00000000000000000000803f0000803f000000000000803f000000000000803f000000000000803f0000803f00000000000000009a99993e9a99993e9a99993e000000009a99193f9a99193f9a99193f000000000000803f0000803f0000803f00000000 \n",
+    "0000803f000000000000000000000000000000000000803f0000000000000000000000000"
+    "00000000000803f00000000000000000000803f0000803f000000000000803f0000000000"
+    "00803f000000000000803f0000803f00000000000000009a99993e9a99993e9a99993e000"
+    "000009a99193F9A99193f9a99193f000000000000803f0000803f0000803f00000000  \n",
+
+    // Test use of storage/buffer resources
+    "//!HOOK MAIN                                                           \n"
+    "//!DESC attach some storage objects                                    \n"
+    "//!BIND tex_storage                                                    \n"
+    "//!BIND buf_uniform                                                    \n"
+    "//!BIND buf_storage                                                    \n"
+    "//!COMPONENTS 4                                                        \n"
+    "                                                                       \n"
+    "vec4 hook()                                                            \n"
+    "{                                                                      \n"
+    "    return vec4(foo, bar, bat);                                        \n"
+    "}                                                                      \n"
+    "                                                                       \n"
+    "//!TEXTURE tex_storage                                                 \n"
+    "//!SIZE 100 100                                                        \n"
+    "//!FORMAT r32f                                                         \n"
+    "//!STORAGE                                                             \n"
+    "                                                                       \n"
+    "//!BUFFER buf_uniform                                                  \n"
+    "//!VAR float foo                                                       \n"
+    "//!VAR float bar                                                       \n"
+    "0000000000000000                                                       \n"
+    "                                                                       \n"
+    "//!BUFFER buf_storage                                                  \n"
+    "//!VAR vec2 bat                                                        \n"
+    "//!VAR int big[32];                                                    \n"
+    "//!STORAGE                                                             \n"
 
 };
 
+static const char *test_luts[] = {
+
+    "TITLE \"1D identity\"  \n"
+    "LUT_1D_SIZE 2          \n"
+    "0.0 0.0 0.0            \n"
+    "1.0 1.0 1.0            \n",
+
+    "TITLE \"3D identity\"  \n"
+    "LUT_3D_SIZE 2          \n"
+    "0.0 0.0 0.0            \n"
+    "1.0 0.0 0.0            \n"
+    "0.0 1.0 0.0            \n"
+    "1.0 1.0 0.0            \n"
+    "0.0 0.0 1.0            \n"
+    "1.0 0.0 1.0            \n"
+    "0.0 1.0 1.0            \n"
+    "1.0 1.0 1.0            \n"
+
+};
+
+static bool frame_passthrough(const struct pl_gpu *gpu, const struct pl_tex **tex,
+                              const struct pl_source_frame *src, struct pl_frame *out_frame)
+{
+    const struct pl_frame *frame = src->frame_data;
+    *out_frame = *frame;
+    return true;
+}
+
+static enum pl_queue_status get_frame_ptr(struct pl_source_frame *out_frame,
+                                          const struct pl_queue_params *qparams)
+{
+    const struct pl_source_frame **pframe = qparams->priv;
+    if (!(*pframe)->frame_data)
+        return QUEUE_EOF;
+
+    *out_frame = *(*pframe)++;
+    return QUEUE_OK;
+}
+
 static void pl_render_tests(const struct pl_gpu *gpu)
 {
-    const struct pl_fmt *fbo_fmt = pl_find_fmt(gpu, PL_FMT_FLOAT, 4, 16, 32,
-                                               PL_FMT_CAP_RENDERABLE |
-                                               PL_FMT_CAP_BLITTABLE);
-    if (!fbo_fmt)
-        return;
+    const struct pl_tex *img5x5_tex = NULL, *fbo = NULL;
+    struct pl_renderer *rr = NULL;
 
     float *fbo_data = NULL;
     static float data_5x5[5][5] = {
@@ -630,10 +877,8 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     };
 
     const int width = 5, height = 5;
-
     struct pl_plane img5x5 = {0};
-    const struct pl_tex *img5x5_tex = NULL;
-    bool ok = pl_upload_plane(gpu, &img5x5, &img5x5_tex, &(struct pl_plane_data) {
+    struct pl_plane_data img5x5_data = {
         .type = PL_FMT_FLOAT,
         .width = width,
         .height = height,
@@ -641,47 +886,39 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         .component_map  = { 0 },
         .pixel_stride = sizeof(float),
         .pixels = &data_5x5,
-    });
+    };
 
-    if (!ok) {
-        pl_tex_destroy(gpu, &img5x5.texture);
+    if (!pl_recreate_plane(gpu, NULL, &fbo, &img5x5_data))
         return;
-    }
 
-    const struct pl_tex *fbo = pl_tex_create(gpu, &(struct pl_tex_params) {
-        .w              = 40,
-        .h              = 40,
-        .format         = fbo_fmt,
-        .renderable     = true,
-        .blit_dst       = true,
-        .storable       = !!(fbo_fmt->caps & PL_FMT_CAP_STORABLE),
-        .host_readable  = true,
-    });
-
-    struct pl_renderer *rr = pl_renderer_create(gpu->ctx, gpu);
-    if (!fbo || !rr)
+    if (!pl_upload_plane(gpu, &img5x5, &img5x5_tex, &img5x5_data))
         goto error;
 
+    rr = pl_renderer_create(gpu->ctx, gpu);
     pl_tex_clear(gpu, fbo, (float[4]){0});
 
-    struct pl_image image = {
-        .signature      = 0,
+    struct pl_frame image = {
         .num_planes     = 1,
         .planes         = { img5x5 },
         .repr = {
             .sys        = PL_COLOR_SYSTEM_BT_709,
-            .levels     = PL_COLOR_LEVELS_PC,
+            .levels     = PL_COLOR_LEVELS_FULL,
         },
         .color          = pl_color_space_bt709,
-        .src_rect       = {-1.0, 0.0, width - 1.0, height},
+        .crop           = {-1.0, 0.0, width - 1.0, height},
     };
 
-    struct pl_render_target target = {
-        .fbo            = fbo,
-        .dst_rect       = {2, 2, fbo->params.w - 2, fbo->params.h - 2},
+    struct pl_frame target = {
+        .num_planes     = 1,
+        .planes         = {{
+            .texture            = fbo,
+            .components         = 3,
+            .component_mapping  = {0, 1, 2},
+        }},
+        .crop           = {2, 2, fbo->params.w - 2, fbo->params.h - 2},
         .repr = {
             .sys        = PL_COLOR_SYSTEM_RGB,
-            .levels     = PL_COLOR_LEVELS_PC,
+            .levels     = PL_COLOR_LEVELS_FULL,
         },
         .color          = pl_color_space_srgb,
     };
@@ -715,9 +952,9 @@ static void pl_render_tests(const struct pl_gpu *gpu)
 #define TEST_PARAMS(NAME, FIELD, LIMIT) \
     TEST(NAME##_params, pl_##NAME##_params, pl_##NAME##_default_params, FIELD, LIMIT)
 
-    for (const struct pl_named_filter_config *f = pl_named_filters; f->name; f++) {
+    for (int i = 0; i < pl_num_filter_presets; i++) {
         struct pl_render_params params = pl_render_default_params;
-        params.upscaler = f->filter;
+        params.upscaler = pl_filter_presets[i].filter;
         REQUIRE(pl_render_image(rr, &image, &target, &params));
         pl_gpu_flush(gpu);
     }
@@ -744,11 +981,12 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         .contrast = 0.9,
         .saturation = 1.5,
         .gamma = 0.8,
+        .temperature = 0.3,
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
     params = pl_render_default_params;
 
-    params.force_3dlut = true;
+    params.force_icc_lut = true;
     REQUIRE(pl_render_image(rr, &image, &target, &params));
     params = pl_render_default_params;
 
@@ -762,7 +1000,14 @@ static void pl_render_tests(const struct pl_gpu *gpu)
         const struct pl_hook *hook;
         hook = pl_mpv_user_shader_parse(gpu, user_shader_tests[i],
                                         strlen(user_shader_tests[i]));
-        REQUIRE(hook);
+
+        if (gpu->caps & PL_GPU_CAP_COMPUTE) {
+            REQUIRE(hook);
+        } else {
+            // Not all shaders compile without compute shader support
+            if (!hook)
+                continue;
+        }
 
         params.hooks = &hook;
         params.num_hooks = 1;
@@ -772,6 +1017,26 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     }
     params = pl_render_default_params;
 
+    // Test custom LUTs
+    for (int i = 0; i < PL_ARRAY_SIZE(test_luts); i++) {
+        printf("testing custom lut %d\n", i);
+        struct pl_custom_lut *lut;
+        lut = pl_lut_parse_cube(gpu->ctx, test_luts[i], strlen(test_luts[i]));
+        REQUIRE(lut);
+
+        // Test all three at the same time to reduce the number of tests
+        image.lut = target.lut = params.lut = lut;
+
+        for (enum pl_lut_type t = PL_LUT_UNKNOWN; t <= PL_LUT_CONVERSION; t++) {
+            printf("testing LUT method %d\n", t);
+            image.lut_type = target.lut_type = params.lut_type = t;
+            REQUIRE(pl_render_image(rr, &image, &target, &params));
+        }
+
+        image.lut = target.lut = params.lut = NULL;
+        pl_lut_free(&lut);
+    }
+
     // Test overlays
     image.num_overlays = 1;
     image.overlays = &(struct pl_overlay) {
@@ -795,6 +1060,77 @@ static void pl_render_tests(const struct pl_gpu *gpu)
     REQUIRE(pl_render_image(rr, &image, &target, &params));
     target.num_overlays = 0;
 
+    // Attempt frame mixing, using the mixer queue helper
+    printf("testing frame mixing \n");
+    struct pl_render_params mix_params = {
+        .frame_mixer = &pl_filter_mitchell_clamp,
+    };
+
+    struct pl_queue_params qparams = {
+        .radius = pl_frame_mix_radius(&mix_params),
+        .vsync_duration = 1.0 / 60.0,
+        .frame_duration = 1.0 / 24.0,
+    };
+
+#define NUM_MIX_FRAMES 20
+    struct pl_source_frame srcframes[NUM_MIX_FRAMES+1];
+    srcframes[NUM_MIX_FRAMES] = (struct pl_source_frame) {0};
+    for (int i = 0; i < NUM_MIX_FRAMES; i++) {
+        srcframes[i] = (struct pl_source_frame) {
+            .pts = i * qparams.frame_duration,
+            .map = frame_passthrough,
+            .frame_data = &image,
+        };
+    }
+
+    struct pl_queue *queue = pl_queue_create(gpu);
+    enum pl_queue_status ret;
+
+    // Test pre-pushing all frames, with delayed EOF
+    for (int i = 0; i < NUM_MIX_FRAMES; i++)
+        pl_queue_push(queue, &srcframes[i]);
+
+    struct pl_frame_mix mix;
+    while ((ret = pl_queue_update(queue, &mix, &qparams)) != QUEUE_EOF) {
+        if (ret == QUEUE_MORE) {
+            REQUIRE(qparams.pts > 0.0);
+            pl_queue_push(queue, NULL); // push delayed EOF
+            continue;
+        }
+
+        REQUIRE(ret == QUEUE_OK);
+        REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
+
+        // Simulate advancing vsync
+        qparams.pts += qparams.vsync_duration;
+    }
+
+    // Test dynamically pulling all frames, with oversample mixer
+    const struct pl_source_frame *frame_ptr = &srcframes[0];
+    mix_params.frame_mixer = &pl_oversample_frame_mixer;
+
+    qparams = (struct pl_queue_params) {
+        .radius = pl_frame_mix_radius(&mix_params),
+        .vsync_duration = qparams.vsync_duration,
+        .frame_duration = qparams.frame_duration,
+        .get_frame = get_frame_ptr,
+        .priv = &frame_ptr,
+    };
+
+    pl_queue_reset(queue);
+    while ((ret = pl_queue_update(queue, &mix, &qparams)) != QUEUE_EOF) {
+        REQUIRE(ret == QUEUE_OK);
+        REQUIRE(mix.num_frames <= 2);
+        REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
+        qparams.pts += qparams.vsync_duration;
+    }
+
+    // Test large PTS jump
+    pl_queue_reset(queue);
+    REQUIRE(pl_queue_update(queue, &mix, &qparams) == QUEUE_EOF);
+
+    pl_queue_destroy(&queue);
+
 error:
     free(fbo_data);
     pl_renderer_destroy(&rr);
@@ -802,11 +1138,260 @@ error:
     pl_tex_destroy(gpu, &fbo);
 }
 
-static void gpu_tests(const struct pl_gpu *gpu)
+static struct pl_hook_res noop_hook(void *priv, const struct pl_hook_params *params)
+{
+    return (struct pl_hook_res) {0};
+}
+
+static void pl_ycbcr_tests(const struct pl_gpu *gpu)
+{
+    struct pl_renderer *rr = pl_renderer_create(gpu->ctx, gpu);
+    if (!rr)
+        return;
+
+    struct pl_plane_data data[3];
+    for (int i = 0; i < 3; i++) {
+        const int sub = i > 0 ? 1 : 0;
+        const int width = (323 + sub) >> sub;
+        const int height = (255 + sub) >> sub;
+
+        data[i] = (struct pl_plane_data) {
+            .type = PL_FMT_UNORM,
+            .width = width,
+            .height = height,
+            .component_size = {16},
+            .component_map = {i},
+            .pixel_stride = sizeof(uint16_t),
+            .row_stride = PL_ALIGN2(width * sizeof(uint16_t),
+                                    gpu->limits.align_tex_xfer_stride),
+        };
+    }
+
+    const struct pl_fmt *fmt = pl_plane_find_fmt(gpu, NULL, &data[0]);
+    if (!fmt || !(fmt->caps & (PL_FMT_CAP_RENDERABLE | PL_FMT_CAP_HOST_READABLE)))
+        return;
+
+    const struct pl_tex *src_tex[3] = {0};
+    const struct pl_tex *dst_tex[3] = {0};
+    struct pl_frame img = {
+        .num_planes = 3,
+        .repr = pl_color_repr_hdtv,
+        .color = pl_color_space_bt709,
+    };
+
+    struct pl_frame target = {
+        .num_planes = 3,
+        .repr = pl_color_repr_hdtv,
+        .color = pl_color_space_bt709,
+    };
+
+    uint8_t *src_buffer[3] = {0};
+    uint8_t *dst_buffer = NULL;
+    for (int i = 0; i < 3; i++) {
+        // Generate some arbitrary data for the buffer
+        src_buffer[i] = malloc(data[i].height * data[i].row_stride);
+        if (!src_buffer[i])
+            goto error;
+
+        data[i].pixels = src_buffer[i];
+        for (int y = 0; y < data[i].height; y++) {
+            for (int x = 0; x < data[i].width; x++) {
+                size_t off = y * data[i].row_stride + x * data[i].pixel_stride;
+                uint16_t *pixel = (uint16_t *) &src_buffer[i][off];
+                int gx = 200 + 100 * i, gy = 300 + 150 * i;
+                *pixel = (gx * x) ^ (gy * y); // whatever
+            }
+        }
+
+        REQUIRE(pl_upload_plane(gpu, &img.planes[i], &src_tex[i], &data[i]));
+    }
+
+    // This co-sites chroma pixels with pixels in the RGB image, meaning we
+    // get an exact round-trip when sampling both ways. This makes it useful
+    // as a test case, even though it's not common in the real world.
+    pl_frame_set_chroma_location(&img, PL_CHROMA_TOP_LEFT);
+
+    for (int i = 0; i < 3; i++) {
+        dst_tex[i] = pl_tex_create(gpu, &(struct pl_tex_params) {
+            .format = fmt,
+            .w = data[i].width,
+            .h = data[i].height,
+            .renderable = true,
+            .host_readable = true,
+            .storable = fmt->caps & PL_FMT_CAP_STORABLE,
+            .blit_dst = fmt->caps & PL_FMT_CAP_BLITTABLE,
+        });
+
+        if (!dst_tex[i])
+            goto error;
+
+        target.planes[i] = img.planes[i];
+        target.planes[i].texture = dst_tex[i];
+    }
+
+    REQUIRE(pl_render_image(rr, &img, &target, &(struct pl_render_params) {
+        .num_hooks = 1,
+        .hooks = &(const struct pl_hook *){&(struct pl_hook) {
+            // Forces chroma merging, to test the chroma merging code
+            .stages = PL_HOOK_CHROMA_INPUT,
+            .hook = noop_hook,
+        }},
+    }));
+
+    size_t buf_size = data[0].height * data[0].row_stride;
+    dst_buffer = malloc(buf_size);
+    if (!dst_buffer)
+        goto error;
+
+    for (int i = 0; i < 3; i++) {
+        memset(dst_buffer, 0xAA, buf_size);
+        REQUIRE(pl_tex_download(gpu, &(struct pl_tex_transfer_params) {
+            .tex = dst_tex[i],
+            .ptr = dst_buffer,
+            .stride_w = data[i].row_stride / data[i].pixel_stride,
+        }));
+
+        for (int y = 0; y < data[i].height; y++) {
+            for (int x = 0; x < data[i].width; x++) {
+                size_t off = y * data[i].row_stride + x * data[i].pixel_stride;
+                uint16_t *src_pixel = (uint16_t *) &src_buffer[i][off];
+                uint16_t *dst_pixel = (uint16_t *) &dst_buffer[off];
+                int diff = abs((int) *src_pixel - (int) *dst_pixel);
+                REQUIRE(diff <= 50); // a little under 0.1%
+            }
+        }
+    }
+
+error:
+    pl_renderer_destroy(&rr);
+    free(dst_buffer);
+    for (int i = 0; i < 3; i++) {
+        free(src_buffer[i]);
+        pl_tex_destroy(gpu, &src_tex[i]);
+        pl_tex_destroy(gpu, &dst_tex[i]);
+    }
+}
+
+static void pl_test_export_import(const struct pl_gpu *gpu,
+                                  enum pl_handle_type handle_type)
+{
+    // Test texture roundtrip
+
+    if (!(gpu->export_caps.tex & handle_type) ||
+        !(gpu->import_caps.tex & handle_type))
+        goto skip_tex;
+
+    const struct pl_fmt *fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0,
+                                           PL_FMT_CAP_BLITTABLE);
+    if (!fmt)
+        goto skip_tex;
+
+    printf("testing texture import/export\n");
+
+    const struct pl_tex *export = pl_tex_create(gpu, &(struct pl_tex_params) {
+        .w = 32,
+        .h = 32,
+        .format = fmt,
+        .export_handle = handle_type,
+    });
+    REQUIRE(export);
+    REQUIRE(export->shared_mem.handle.fd > -1);
+
+    const struct pl_tex *import = pl_tex_create(gpu, &(struct pl_tex_params) {
+        .w = 32,
+        .h = 32,
+        .format = fmt,
+        .import_handle = handle_type,
+        .shared_mem = export->shared_mem,
+    });
+    REQUIRE(import);
+
+    pl_tex_destroy(gpu, &import);
+    pl_tex_destroy(gpu, &export);
+
+skip_tex: ;
+
+    // Test buffer roundtrip
+
+    if (!(gpu->export_caps.buf & handle_type) ||
+        !(gpu->import_caps.buf & handle_type))
+        return;
+
+    printf("testing buffer import/export\n");
+
+    const struct pl_buf *exp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .size = 32,
+        .export_handle = handle_type,
+    });
+    REQUIRE(exp_buf);
+    REQUIRE(exp_buf->shared_mem.handle.fd > -1);
+
+    const struct pl_buf *imp_buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .size = 32,
+        .import_handle = handle_type,
+        .shared_mem = exp_buf->shared_mem,
+    });
+    REQUIRE(imp_buf);
+
+    pl_buf_destroy(gpu, &imp_buf);
+    pl_buf_destroy(gpu, &exp_buf);
+}
+
+static void pl_test_host_ptr(const struct pl_gpu *gpu)
+{
+    if (!(gpu->import_caps.buf & PL_HANDLE_HOST_PTR))
+        return;
+
+#ifdef __unix__
+
+    printf("testing host ptr\n");
+    REQUIRE(gpu->caps & PL_GPU_CAP_MAPPED_BUFFERS);
+
+    const size_t size = 2 << 20;
+    const size_t offset = 2 << 10;
+    const size_t slice = 2 << 16;
+
+    uint8_t *data = aligned_alloc(0x1000, size);
+    for (int i = 0; i < size; i++)
+        data[i] = (uint8_t) i;
+
+    const struct pl_buf *buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        .type = PL_BUF_TEX_TRANSFER,
+        .size = slice,
+        .import_handle = PL_HANDLE_HOST_PTR,
+        .shared_mem = {
+            .handle.ptr = data,
+            .size = size,
+            .offset = offset,
+        },
+        .host_mapped = true,
+    });
+
+    REQUIRE(buf);
+    REQUIRE(memcmp(data + offset, buf->data, slice) == 0);
+
+    pl_buf_destroy(gpu, &buf);
+    free(data);
+
+#endif // unix
+}
+
+static void gpu_shader_tests(const struct pl_gpu *gpu)
 {
     pl_buffer_tests(gpu);
     pl_texture_tests(gpu);
     pl_shader_tests(gpu);
     pl_scaler_tests(gpu);
     pl_render_tests(gpu);
+    pl_ycbcr_tests(gpu);
+
+    REQUIRE(!pl_gpu_is_failed(gpu));
+}
+
+static void gpu_interop_tests(const struct pl_gpu *gpu)
+{
+    pl_test_export_import(gpu, PL_HANDLE_DMA_BUF);
+    pl_test_host_ptr(gpu);
+
+    REQUIRE(!pl_gpu_is_failed(gpu));
 }
