@@ -7,12 +7,12 @@ static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
                                  enum pl_handle_type handle_type)
 {
     const struct pl_gpu *gpu = pl_vk->gpu;
+    printf("testing vulkan interop for handle type 0x%x\n", handle_type);
 
     if (gpu->export_caps.buf & handle_type) {
         const struct pl_buf *buf = pl_buf_create(gpu, &(struct pl_buf_params) {
-            .type = PL_BUF_TEX_TRANSFER,
             .size = 1024,
-            .handle_type = handle_type,
+            .export_handle = handle_type,
         });
 
         REQUIRE(buf);
@@ -47,7 +47,7 @@ static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
         REQUIRE(pl_tex_export(gpu, tex, sync));
 
         // Re-use our internal helpers to signal this VkSemaphore
-        struct vk_ctx *vk = TA_PRIV(pl_vk);
+        struct vk_ctx *vk = PL_PRIV(pl_vk);
         struct vk_cmd *cmd = vk_cmd_begin(vk, vk->pool_graphics);
         VkSemaphore signal;
         REQUIRE(cmd);
@@ -59,53 +59,19 @@ static void vulkan_interop_tests(const struct pl_vulkan *pl_vk,
         // Do something with the image again to "import" it
         pl_tex_clear(gpu, tex, (float[4]){0});
         pl_gpu_finish(gpu);
+        REQUIRE(!pl_tex_poll(gpu, tex, 0));
 
         pl_sync_destroy(gpu, &sync);
         pl_tex_destroy(gpu, &tex);
     }
 }
 
-static void vulkan_test_export_import(const struct pl_vulkan *pl_vk,
-                                      enum pl_handle_type handle_type)
-{
-    const struct pl_gpu *gpu = pl_vk->gpu;
-
-    if (!(gpu->export_caps.tex & handle_type) ||
-        !(gpu->import_caps.tex & handle_type))
-        return;
-
-    const struct pl_fmt *fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0,
-                                           PL_FMT_CAP_BLITTABLE);
-    if (!fmt)
-        return;
-
-    const struct pl_tex *export = pl_tex_create(gpu, &(struct pl_tex_params) {
-        .w = 32,
-        .h = 32,
-        .format = fmt,
-        .export_handle = handle_type,
-    });
-    REQUIRE(export);
-    REQUIRE(export->shared_mem.handle.fd > -1);
-
-    const struct pl_tex *import = pl_tex_create(gpu, &(struct pl_tex_params) {
-        .w = 32,
-        .h = 32,
-        .format = fmt,
-        .import_handle = handle_type,
-        .shared_mem = export->shared_mem,
-    });
-    REQUIRE(import);
-
-    pl_tex_destroy(gpu, &import);
-    pl_tex_destroy(gpu, &export);
-}
-
 static void vulkan_swapchain_tests(const struct pl_vulkan *vk, VkSurfaceKHR surf)
 {
     if (!surf)
         return;
 
+    printf("testing vulkan swapchain\n");
     const struct pl_gpu *gpu = vk->gpu;
     const struct pl_swapchain *sw;
     sw = pl_vulkan_create_swapchain(vk, &(struct pl_vulkan_swapchain_params) {
@@ -124,8 +90,8 @@ static void vulkan_swapchain_tests(const struct pl_vulkan *vk, VkSurfaceKHR surf
             pl_tex_clear(gpu, frame.fbo, (float[4]){0});
 
         // TODO: test this with an actual pl_renderer instance
-        struct pl_render_target target;
-        pl_render_target_from_swapchain(&target, &frame);
+        struct pl_frame target;
+        pl_frame_from_swapchain(&target, &frame);
 
         REQUIRE(pl_swapchain_submit_frame(sw));
         pl_swapchain_swap_buffers(sw);
@@ -158,8 +124,8 @@ int main()
     if (!inst)
         return SKIP;
 
-    VK_LOAD_FUN(inst->instance, EnumeratePhysicalDevices, vkGetInstanceProcAddr);
-    VK_LOAD_FUN(inst->instance, GetPhysicalDeviceProperties, vkGetInstanceProcAddr);
+    PL_VK_LOAD_FUN(inst->instance, EnumeratePhysicalDevices, vkGetInstanceProcAddr);
+    PL_VK_LOAD_FUN(inst->instance, GetPhysicalDeviceProperties, vkGetInstanceProcAddr);
 
     uint32_t num = 0;
     EnumeratePhysicalDevices(inst->instance, &num, NULL);
@@ -174,7 +140,7 @@ int main()
     VkSurfaceKHR surf = NULL;
 
 #ifdef VK_EXT_headless_surface
-    VK_LOAD_FUN(inst->instance, CreateHeadlessSurfaceEXT, vkGetInstanceProcAddr);
+    PL_VK_LOAD_FUN(inst->instance, CreateHeadlessSurfaceEXT, vkGetInstanceProcAddr);
     if (CreateHeadlessSurfaceEXT) {
         VkHeadlessSurfaceCreateInfoEXT info = {
             .sType = VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT,
@@ -213,15 +179,11 @@ int main()
         params.queue_count = 8; // test inter-queue stuff
         params.surface = surf;
 
-#ifdef CI_BLACKLIST_COMPUTE
-        params.blacklist_caps = PL_GPU_CAP_COMPUTE;
-#endif
-
         const struct pl_vulkan *vk = pl_vulkan_create(ctx, &params);
         if (!vk)
             continue;
 
-        gpu_tests(vk->gpu);
+        gpu_shader_tests(vk->gpu);
         vulkan_swapchain_tests(vk, surf);
 
         // Test importing this context via the vulkan interop API
@@ -242,16 +204,15 @@ int main()
         pl_vulkan_destroy(&vk2);
 
         // Run these tests last because they disable some validation layers
-#ifdef VK_HAVE_UNIX
+#ifdef PL_HAVE_UNIX
         vulkan_interop_tests(vk, PL_HANDLE_FD);
         vulkan_interop_tests(vk, PL_HANDLE_DMA_BUF);
-        vulkan_test_export_import(vk, PL_HANDLE_DMA_BUF);
 #endif
-#ifdef VK_HAVE_WIN32
+#ifdef PL_HAVE_WIN32
         vulkan_interop_tests(vk, PL_HANDLE_WIN32);
         vulkan_interop_tests(vk, PL_HANDLE_WIN32_KMT);
 #endif
-
+        gpu_interop_tests(vk->gpu);
         pl_vulkan_destroy(&vk);
 
         // Re-run the same export/import tests with async queues disabled
@@ -260,16 +221,15 @@ int main()
         vk = pl_vulkan_create(ctx, &params);
         REQUIRE(vk); // it succeeded the first time
 
-#ifdef VK_HAVE_UNIX
+#ifdef PL_HAVE_UNIX
         vulkan_interop_tests(vk, PL_HANDLE_FD);
         vulkan_interop_tests(vk, PL_HANDLE_DMA_BUF);
-        vulkan_test_export_import(vk, PL_HANDLE_DMA_BUF);
 #endif
-#ifdef VK_HAVE_WIN32
+#ifdef PL_HAVE_WIN32
         vulkan_interop_tests(vk, PL_HANDLE_WIN32);
         vulkan_interop_tests(vk, PL_HANDLE_WIN32_KMT);
 #endif
-
+        gpu_interop_tests(vk->gpu);
         pl_vulkan_destroy(&vk);
 
         // Reduce log spam after first tested device
