@@ -22,6 +22,11 @@
 #include "gpu.h"
 #include "swapchain.h"
 
+struct sem_pair {
+    VkSemaphore in;
+    VkSemaphore out;
+};
+
 struct priv {
     struct vk_ctx *vk;
     VkSurfaceKHR surf;
@@ -40,13 +45,10 @@ struct priv {
     struct pl_hdr_metadata hdr_metadata;
 
     // state of the images:
-    const struct pl_tex **images; // pl_tex wrappers for the VkImages
-    int num_images;         // size of `images`
-    VkSemaphore *sems_in;   // pool of semaphores used to synchronize images
-    VkSemaphore *sems_out;  // outgoing semaphores (rendering complete)
-    int num_sems;           // size of `sems_in` / `sems_out`
-    int idx_sems;           // index of next free semaphore pair
-    int last_imgidx;        // the image index last acquired (for submit)
+    PL_ARRAY(const struct pl_tex *) images; // pl_tex wrappers for the VkImages
+    PL_ARRAY(struct sem_pair) sems;         // pool of semaphores used to synchronize images
+    int idx_sems;                           // index of next free semaphore pair
+    int last_imgidx;                        // the image index last acquired (for submit)
 };
 
 static struct pl_sw_fns vulkan_swapchain;
@@ -143,15 +145,20 @@ static bool vk_map_color_space(VkColorSpaceKHR space, struct pl_color_space *out
 }
 
 static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
-                             VkSurfaceKHR surf, VkSurfaceFormatKHR *out_format,
-                             struct pl_color_space *space)
+                             VkSurfaceKHR surf, bool prefer_hdr,
+                             VkSurfaceFormatKHR *out_format,
+                             struct pl_color_space *out_space)
 {
+    int best_score = 0, best_id;
     VkSurfaceFormatKHR *formats = NULL;
     int num = 0;
 
     // Specific format requested by user
     if (out_format->format) {
-        if (vk_map_color_space(out_format->colorSpace, space)) {
+        if (vk_map_color_space(out_format->colorSpace, out_space)) {
+            PL_INFO(gpu, "Using user-supplied surface configuration: %s + %s",
+                    vk_fmt_name(out_format->format),
+                    vk_csp_name(out_format->colorSpace));
             return true;
         } else {
             PL_ERR(gpu, "User-supplied surface format unsupported: %s",
@@ -160,14 +167,14 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
     }
 
     VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, NULL));
-    formats = talloc_zero_array(NULL, VkSurfaceFormatKHR, num);
+    formats = pl_calloc_ptr(NULL, num, formats);
     VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, surf, &num, formats));
 
-    PL_DEBUG(gpu, "Available surface formats:");
+    PL_INFO(gpu, "Available surface configurations:");
     for (int i = 0; i < num; i++) {
-        PL_DEBUG(gpu, "    %d: format: %s, space: %s", i,
-                 vk_fmt_name(formats[i].format),
-                 vk_csp_name(formats[i].colorSpace));
+        PL_INFO(gpu, "    %d: %-40s %s", i,
+                vk_fmt_name(formats[i].format),
+                vk_csp_name(formats[i].colorSpace));
     }
 
     for (int i = 0; i < num; i++) {
@@ -177,12 +184,13 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
                 .colorSpace = VK_COLOR_SPACE_SRGB_NONLINEAR_KHR,
                 .format = VK_FORMAT_R8G8B8A8_UNORM,
             };
-            talloc_free(formats);
+            pl_free(formats);
             return true;
         }
 
         // Color space / format whitelist
-        if (!vk_map_color_space(formats[i].colorSpace, space))
+        struct pl_color_space space;
+        if (!vk_map_color_space(formats[i].colorSpace, &space))
             continue;
 
         switch (formats[i].format) {
@@ -193,7 +201,7 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
         case VK_FORMAT_R32G32B32A32_SFLOAT:
         case VK_FORMAT_R64G64B64_SFLOAT:
         case VK_FORMAT_R64G64B64A64_SFLOAT:
-            if (space->transfer == PL_COLOR_TRC_LINEAR)
+            if (space.transfer == PL_COLOR_TRC_LINEAR)
                 break; // accept
             continue;
 
@@ -203,14 +211,14 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
         case VK_FORMAT_R8G8B8A8_UNORM:
         case VK_FORMAT_B8G8R8A8_UNORM:
         case VK_FORMAT_A8B8G8R8_UNORM_PACK32:
-            if (!pl_color_transfer_is_hdr(space->transfer))
+            if (!pl_color_transfer_is_hdr(space.transfer))
                 break; // accept
             continue;
 
         // Only accept 10 bit formats for non-linear curves
         case VK_FORMAT_A2R10G10B10_UNORM_PACK32:
         case VK_FORMAT_A2B10G10R10_UNORM_PACK32:
-            if (space->transfer != PL_COLOR_TRC_LINEAR)
+            if (space.transfer != PL_COLOR_TRC_LINEAR)
                 break; // accept
             continue;
 
@@ -225,7 +233,7 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
         // Make sure we can wrap this format to a meaningful, valid pl_fmt
         for (int n = 0; n < gpu->num_formats; n++) {
             const struct pl_fmt *plfmt = gpu->formats[n];
-            const struct vk_format **pvkfmt = TA_PRIV(plfmt);
+            const struct vk_format **pvkfmt = PL_PRIV(plfmt);
             if ((*pvkfmt)->tfmt != formats[i].format)
                 continue;
 
@@ -235,24 +243,45 @@ static bool pick_surf_format(const struct pl_gpu *gpu, const struct vk_ctx *vk,
             if ((plfmt->caps & render_caps) != render_caps)
                 continue;
 
-            // format valid, use it
-            *out_format = formats[i];
-            talloc_free(formats);
-            return true;
+            // format valid, use it if it has a higher score
+            int score = 0;
+            for (int c = 0; c < 3; c++)
+                score += plfmt->component_depth[c];
+            if (pl_color_transfer_is_hdr(space.transfer) == prefer_hdr)
+                score += 10000;
+
+            switch (plfmt->type) {
+            case PL_FMT_UNORM: score += 3000; break;
+            case PL_FMT_SNORM: score += 2000; break;
+            case PL_FMT_FLOAT: score += 1000; break;
+            default: break;
+            };
+
+            if (score > best_score) {
+                *out_format = formats[i];
+                *out_space = space;
+                best_score = score;
+                best_id = i;
+                break;
+            }
         }
     }
 
+    if (best_score)
+        PL_INFO(gpu, "Picked surface configuration %d", best_id);
+
     // fall through
 error:
-    PL_FATAL(vk, "Failed picking any valid, renderable surface format!");
-    talloc_free(formats);
-    return false;
+    if (!best_score)
+        PL_FATAL(vk, "Failed picking any valid, renderable surface format!");
+    pl_free(formats);
+    return best_score > 0;
 }
 
 const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *plvk,
                               const struct pl_vulkan_swapchain_params *params)
 {
-    struct vk_ctx *vk = TA_PRIV(plvk);
+    struct vk_ctx *vk = PL_PRIV(plvk);
     const struct pl_gpu *gpu = plvk->gpu;
 
     if (!vk->CreateSwapchainKHR) {
@@ -262,18 +291,15 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
 
     VkSurfaceFormatKHR sfmt = params->surface_format;
     struct pl_color_space csp;
-    if (!pick_surf_format(gpu, vk, params->surface, &sfmt, &csp))
+    if (!pick_surf_format(gpu, vk, params->surface, params->prefer_hdr, &sfmt, &csp))
         return NULL;
 
-    PL_DEBUG(gpu, "Picked surface format %s, space %s",
-             vk_fmt_name(sfmt.format), vk_csp_name(sfmt.colorSpace));
-
-    struct pl_swapchain *sw = talloc_zero_priv(NULL, struct pl_swapchain, struct priv);
+    struct pl_swapchain *sw = pl_zalloc_priv(NULL, struct pl_swapchain, struct priv);
     sw->impl = &vulkan_swapchain;
     sw->ctx = vk->ctx;
     sw->gpu = gpu;
 
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     p->params = *params;
     p->vk = vk;
     p->surf = params->surface;
@@ -294,23 +320,21 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
     p->color_space = csp;
     p->color_repr = (struct pl_color_repr) {
         .sys    = PL_COLOR_SYSTEM_RGB,
-        .levels = PL_COLOR_LEVELS_PC,
+        .levels = PL_COLOR_LEVELS_FULL,
         .alpha  = PL_ALPHA_UNKNOWN, // will be updated by vk_sw_recreate
     };
 
     // Make sure the swapchain present mode is supported
     VkPresentModeKHR *modes = NULL;
     int num_modes = 0;
-    VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf,
-                                                   &num_modes, NULL));
-    modes = talloc_zero_array(NULL, VkPresentModeKHR, num_modes);
-    VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf,
-                                                   &num_modes, modes));
+    VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf, &num_modes, NULL));
+    modes = pl_calloc_ptr(NULL, num_modes, modes);
+    VK(vk->GetPhysicalDeviceSurfacePresentModesKHR(vk->physd, p->surf, &num_modes, modes));
 
     bool supported = false;
     for (int i = 0; i < num_modes; i++)
         supported |= (modes[i] == p->protoInfo.presentMode);
-    TA_FREEP(&modes);
+    pl_free_ptr(&modes);
 
     if (!supported) {
         PL_WARN(vk, "Requested swap mode unsupported by this device, falling "
@@ -321,33 +345,33 @@ const struct pl_swapchain *pl_vulkan_create_swapchain(const struct pl_vulkan *pl
     return sw;
 
 error:
-    talloc_free(modes);
-    talloc_free(sw);
+    pl_free(modes);
+    pl_free(sw);
     return NULL;
 }
 
 static void vk_sw_destroy(const struct pl_swapchain *sw)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
     pl_gpu_flush(gpu);
     vk_wait_idle(vk);
-    for (int i = 0; i < p->num_images; i++)
-        pl_tex_destroy(gpu, &p->images[i]);
-    for (int i = 0; i < p->num_sems; i++) {
-        vk->DestroySemaphore(vk->dev, p->sems_in[i], VK_ALLOC);
-        vk->DestroySemaphore(vk->dev, p->sems_out[i], VK_ALLOC);
+    for (int i = 0; i < p->images.num; i++)
+        pl_tex_destroy(gpu, &p->images.elem[i]);
+    for (int i = 0; i < p->sems.num; i++) {
+        vk->DestroySemaphore(vk->dev, p->sems.elem[i].in, PL_VK_ALLOC);
+        vk->DestroySemaphore(vk->dev, p->sems.elem[i].out, PL_VK_ALLOC);
     }
 
-    vk->DestroySwapchainKHR(vk->dev, p->swapchain, VK_ALLOC);
-    talloc_free((void *) sw);
+    vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
+    pl_free((void *) sw);
 }
 
 static int vk_sw_latency(const struct pl_swapchain *sw)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     return p->swapchain_depth;
 }
 
@@ -445,15 +469,18 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     }
 
     // Clamp the extent based on the supported limits
-    w = PL_MIN(PL_MAX(w, caps.minImageExtent.width), caps.maxImageExtent.width);
-    h = PL_MIN(PL_MAX(h, caps.minImageExtent.height), caps.maxImageExtent.height);
+    w = PL_CLAMP(w, caps.minImageExtent.width,  caps.maxImageExtent.width);
+    h = PL_CLAMP(h, caps.minImageExtent.height, caps.maxImageExtent.height);
     info->imageExtent = (VkExtent2D) { w, h };
 
-    // We just request whatever we can, and let the pl_vk decide what
-    // pl_tex_params that translates to. This makes the image as flexible as
-    // possible. That said, we still need to intersect the swapchain supported
-    // usage flags with the format supported usage flags
-    info->imageUsage = caps.supportedUsageFlags;
+    // We just request whatever makes sense, and let the pl_vk decide what
+    // pl_tex_params that translates to. That said, we still need to intersect
+    // the swapchain usage flags with the format usage flags
+    VkImageUsageFlags req_flags = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT |
+                                  VK_IMAGE_USAGE_TRANSFER_DST_BIT;
+    VkImageUsageFlags opt_flags = VK_IMAGE_USAGE_STORAGE_BIT;
+
+    info->imageUsage = caps.supportedUsageFlags & (req_flags | opt_flags);
     VkFormatProperties fmtprop = {0};
     vk->GetPhysicalDeviceFormatProperties(vk->physd, info->imageFormat, &fmtprop);
 
@@ -461,17 +488,11 @@ static bool update_swapchain_info(struct priv *p, VkSwapchainCreateInfoKHR *info
     if (!((fmtprop.optimalTilingFeatures & VK_FORMAT_FEATURE_##feature##_BIT))) \
         info->imageUsage &= ~VK_IMAGE_USAGE_##usage##_BIT
 
-    CHECK(TRANSFER_SRC, TRANSFER_SRC);
-    CHECK(TRANSFER_DST, TRANSFER_DST);
-    CHECK(SAMPLED, SAMPLED_IMAGE);
-    CHECK(STORAGE, STORAGE_IMAGE);
     CHECK(COLOR_ATTACHMENT, COLOR_ATTACHMENT);
+    CHECK(TRANSFER_DST, TRANSFER_DST);
+    CHECK(STORAGE, STORAGE_IMAGE);
 
-    // Make sure we support at least rendering and blitting!
-    VkImageUsageFlags required_flags = 0;
-    required_flags |= VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
-    required_flags |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;
-    if ((info->imageUsage & required_flags) != required_flags) {
+    if ((info->imageUsage & req_flags) != req_flags) {
         PL_ERR(vk, "The swapchain doesn't support rendering and blitting!");
         goto error;
     }
@@ -485,14 +506,14 @@ error:
 static void destroy_swapchain(struct vk_ctx *vk, struct priv *p)
 {
     assert(p->old_swapchain);
-    vk->DestroySwapchainKHR(vk->dev, p->old_swapchain, VK_ALLOC);
+    vk->DestroySwapchainKHR(vk->dev, p->old_swapchain, PL_VK_ALLOC);
     p->old_swapchain = VK_NULL_HANDLE;
 }
 
 static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
 
     VkImage *vkimages = NULL;
@@ -516,7 +537,7 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
             sinfo.imageExtent.width,
             sinfo.imageExtent.height);
 
-    VK(vk->CreateSwapchainKHR(vk->dev, &sinfo, VK_ALLOC, &p->swapchain));
+    VK(vk->CreateSwapchainKHR(vk->dev, &sinfo, PL_VK_ALLOC, &p->swapchain));
 
     p->suboptimal = false;
     p->cur_width = sinfo.imageExtent.width;
@@ -531,48 +552,46 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
 
     // Get the new swapchain images
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, NULL));
-    vkimages = talloc_zero_array(NULL, VkImage, num_images);
+    vkimages = pl_calloc_ptr(NULL, num_images, vkimages);
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, vkimages));
 
     for (int i = 0; i < num_images; i++)
-        VK_NAME(IMAGE, vkimages[i], "swapchain");
+        PL_VK_NAME(IMAGE, vkimages[i], "swapchain");
 
     // If needed, allocate some more semaphores
-    while (num_images > p->num_sems) {
+    while (num_images > p->sems.num) {
         VkSemaphore sem_in = NULL, sem_out = NULL;
         static const VkSemaphoreCreateInfo seminfo = {
             .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
         };
-        VK(vk->CreateSemaphore(vk->dev, &seminfo, VK_ALLOC, &sem_in));
-        VK(vk->CreateSemaphore(vk->dev, &seminfo, VK_ALLOC, &sem_out));
+        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem_in));
+        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem_out));
+        PL_VK_NAME(SEMAPHORE, sem_in, "swapchain in");
+        PL_VK_NAME(SEMAPHORE, sem_out, "swapchain out");
 
-        VK_NAME(SEMAPHORE, sem_in, "swapchain in");
-        VK_NAME(SEMAPHORE, sem_out, "swapchain out");
-
-        int idx = p->num_sems++;
-        TARRAY_GROW(sw, p->sems_in, idx);
-        TARRAY_GROW(sw, p->sems_out, idx);
-        p->sems_in[idx] = sem_in;
-        p->sems_out[idx] = sem_out;
+        PL_ARRAY_APPEND(sw, p->sems, (struct sem_pair) {
+            .in = sem_in,
+            .out = sem_out,
+        });
     }
 
     // Recreate the pl_tex wrappers
-    for (int i = 0; i < p->num_images; i++)
-        pl_tex_destroy(gpu, &p->images[i]);
+    for (int i = 0; i < p->images.num; i++)
+        pl_tex_destroy(gpu, &p->images.elem[i]);
+    p->images.num = 0;
 
-    p->num_images = num_images;
-    TARRAY_GROW(sw, p->images, num_images);
     for (int i = 0; i < num_images; i++) {
         const VkExtent2D *ext = &sinfo.imageExtent;
-        p->images[i] = pl_vulkan_wrap(gpu, &(struct pl_vulkan_wrap_params) {
+        const struct pl_tex *tex = pl_vulkan_wrap(gpu, &(struct pl_vulkan_wrap_params) {
             .image = vkimages[i],
             .width = ext->width,
             .height = ext->height,
             .format = sinfo.imageFormat,
             .usage = sinfo.imageUsage,
         });
-        if (!p->images[i])
+        if (!tex)
             goto error;
+        PL_ARRAY_APPEND(sw, p->images, tex);
     }
 
     pl_assert(num_images > 0);
@@ -582,7 +601,7 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     // the actual color information (consider e.g. a2bgr10). Slight downside
     // in that it results in rounding r/b for e.g. rgb565, but we don't pick
     // surfaces with fewer than 8 bits anyway, so let's not care for now.
-    const struct pl_fmt *fmt = p->images[0]->params.format;
+    const struct pl_fmt *fmt = p->images.elem[0]->params.format;
     for (int i = 0; i < fmt->num_components; i++)
         bits = PL_MAX(bits, fmt->component_depth[i]);
 
@@ -592,14 +611,14 @@ static bool vk_sw_recreate(const struct pl_swapchain *sw, int w, int h)
     // Re-set the HDR metadata
     pl_swapchain_hdr_metadata(sw, &p->hdr_metadata);
 
-    talloc_free(vkimages);
+    pl_free(vkimages);
     return true;
 
 error:
     PL_ERR(vk, "Failed (re)creating swapchain!");
-    talloc_free(vkimages);
+    pl_free(vkimages);
     if (p->swapchain != sinfo.oldSwapchain) {
-        vk->DestroySwapchainKHR(vk->dev, p->swapchain, VK_ALLOC);
+        vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
         p->swapchain = VK_NULL_HANDLE;
         p->cur_width = p->cur_height = 0;
         p->suboptimal = false;
@@ -610,7 +629,7 @@ error:
 static bool vk_sw_start_frame(const struct pl_swapchain *sw,
                               struct pl_swapchain_frame *out_frame)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
     if (!p->swapchain && !vk_sw_recreate(sw, 0, 0))
         return false;
@@ -621,7 +640,7 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
             return false;
     }
 
-    VkSemaphore sem_in = p->sems_in[p->idx_sems];
+    VkSemaphore sem_in = p->sems.elem[p->idx_sems].in;
     PL_TRACE(vk, "vkAcquireNextImageKHR signals %p", (void *) sem_in);
 
     for (int attempts = 0; attempts < 2; attempts++) {
@@ -635,10 +654,10 @@ static bool vk_sw_start_frame(const struct pl_swapchain *sw,
             // fall through
         case VK_SUCCESS:
             p->last_imgidx = imgidx;
-            pl_vulkan_release(sw->gpu, p->images[imgidx],
+            pl_vulkan_release(sw->gpu, p->images.elem[imgidx],
                               VK_IMAGE_LAYOUT_UNDEFINED, 0, sem_in);
             *out_frame = (struct pl_swapchain_frame) {
-                .fbo = p->images[imgidx],
+                .fbo = p->images.elem[imgidx],
                 .flipped = false,
                 .color_repr = p->color_repr,
                 .color_space = p->color_space,
@@ -671,15 +690,15 @@ static void present_cb(struct priv *p, void *arg)
 static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
 {
     const struct pl_gpu *gpu = sw->gpu;
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
     if (!p->swapchain)
         return false;
 
-    VkSemaphore sem_out = p->sems_out[p->idx_sems++];
-    p->idx_sems %= p->num_sems;
+    VkSemaphore sem_out = p->sems.elem[p->idx_sems++].out;
+    p->idx_sems %= p->sems.num;
 
-    bool held = pl_vulkan_hold(gpu, p->images[p->last_imgidx],
+    bool held = pl_vulkan_hold(gpu, p->images.elem[p->last_imgidx],
                                VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
                                VK_ACCESS_MEMORY_READ_BIT, sem_out);
     if (!held) {
@@ -735,7 +754,7 @@ static bool vk_sw_submit_frame(const struct pl_swapchain *sw)
 
 static void vk_sw_swap_buffers(const struct pl_swapchain *sw)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
 
     while (p->frames_in_flight >= p->swapchain_depth)
         vk_poll_commands(p->vk, UINT64_MAX);
@@ -743,7 +762,7 @@ static void vk_sw_swap_buffers(const struct pl_swapchain *sw)
 
 static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     bool ok = true;
 
     bool width_changed = *width && *width != p->cur_width,
@@ -760,7 +779,7 @@ static bool vk_sw_resize(const struct pl_swapchain *sw, int *width, int *height)
 static bool vk_sw_hdr_metadata(const struct pl_swapchain *sw,
                                const struct pl_hdr_metadata *metadata)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     struct vk_ctx *vk = p->vk;
     if (!vk->SetHdrMetadataEXT)
         return false;
@@ -794,7 +813,7 @@ static bool vk_sw_hdr_metadata(const struct pl_swapchain *sw,
 
 bool pl_vulkan_swapchain_suboptimal(const struct pl_swapchain *sw)
 {
-    struct priv *p = TA_PRIV(sw);
+    struct priv *p = PL_PRIV(sw);
     return p->suboptimal;
 }
 
